# -*- coding: utf-8 -*-
"""
Created on Mon Jun  3 12:09:55 2019

@author: alxgr
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot      = manipulator.TwoLinkManipulator()

# Settting

#torque_controlled_robot.l1  = 0.5
#torque_controlled_robot.l2  = 0.5
#torque_controlled_robot.lc1 = 0.5
#torque_controlled_robot.lc2 = 0.5

torque_controlled_robot.I1     = 0  
torque_controlled_robot.I2     = 0

torque_controlled_robot.gravity = 9
torque_controlled_robot.d1      = 0.1
torque_controlled_robot.d2      = 0.1
torque_controlled_robot.ubar    = [0,0] 

#torque_controlled_robot.ubar    = [0.5,0] 
#torque_controlled_robot.gravity = 9.8
#torque_controlled_robot.d1      = 0.5
#torque_controlled_robot.d2      = 1


# Simulations

torque_controlled_robot.x0 = np.array([0,0.1,0,0])

torque_controlled_robot.animate_simulation()
#torque_controlled_robot.plot_trajectory('xu')


# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""
import numpy as np

from pyro.dynamic               import massspringdamper
from pyro.control.linear        import PIDController


# Plant
sys = massspringdamper.TwoMass()

sys.m1 = 2
sys.m2 = 3
sys.k1 = 5
sys.k2 = 2
sys.b1 = 0.1
sys.b2 = 0.2

sys.compute_ABCD()


# LQR controller
kp  = 5
ki  = 1
kd  = 2
tau = 0.1

ctl = PIDController(kp, ki, kd, tau)

# Simulation Closed-Loop Non-linear with LQR controller
cl_sys = ctl + sys
cl_sys.x0[1] = 1
cl_sys.compute_trajectory()
cl_sys.plot_trajectory('xu')
cl_sys.plot_linearized_pz_map()
cl_sys.plot_linearized_bode()
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 23 13:40:47 2022

@author: alex
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import vehicle_propulsion
from pyro.planning import discretizer
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming
from pyro.control  import controller
###############################################################################

sys  = vehicle_propulsion.LongitudinalFrontWheelDriveCarWithWheelSlipInput()

###############################################################################

# Planning

# Set domain
sys.x_ub = np.array([+50, 20,])
sys.x_lb = np.array([0, 0])
sys.u_ub = np.array([0.0])
sys.u_lb = np.array([-0.1])

# Discrete world
grid_sys = discretizer.GridDynamicSystem(sys, (101, 101), (11,), 0.1)

# Cost Function
cf = costfunction.QuadraticCostFunction.from_sys( sys )
cf.xbar = np.array( [0, 0] ) # target
cf.INF  = 1E3
cf.EPS  = 0.00
cf.R[0]   = 1
cf.Q[0,0] = 0
cf.Q[1,1] = 0.01

# VI algo
vi = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys , cf )

vi.solve_bellman_equation()

# Closed-loop Law

vi.plot_cost2go_3D()

ctl = vi.get_lookup_table_controller()

cl_sys = ctl + sys

###############################################################################

## Simulation and animation

x0   = np.array([0, 16])
tf   = 10

cl_sys.x0 = x0
cl_sys.compute_trajectory(tf, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation( time_factor_video = 3 )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug  1 22:54:02 2019

@author: alex
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control import controller
###############################################################################


class kinematicInputs( controller.StaticController ) :
    """ 
    Simple proportionnal compensator
    ----------------------------------------
    r  : reference signal_proc vector  k x 1
    y  : sensor signal_proc vector     k x 1
    u  : control inputs vector         k x 1
    t  : time                          1 x 1
    ----------------------------------------
    u = c( y , r , t ) = (r - y) * gain

    """
    
    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################
    
    
    ############################
    def __init__(self, k = 1):
        """ """
        
        # Dimensions
        self.k = 3   
        self.m = 2   
        self.p = 3
        
        super().__init__(self.k, self.m, self.p)
        
        # Label
        self.name = 'Proportionnal Controller'
        
        # Gains
        self.gain = 1
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        max_steer = 0.15
        max_speed = 5.0000
        u = np.zeros(self.m) # State derivative vector
        
        if t<1:
            u[0] = 0
            #u[0] = max_steer
            u[1] = max_speed
            #u[1] = 0
        elif(t>=1 and t<=3):
            u[0] = float(max_steer/2.0000*(t-1.0000))
            #u[0] = max_steer
            #u[1] = float(max_speed/40.0000*(t-10.0000))
            u[1] = max_speed
        elif(t>3 and t<=90):
            #u[0] = float(max_steer-max_steer/40.0000*(t-50.0000))
            u[0] = max_steer
            u[1] = max_speed
        else:
            #u[0] = 0
            u[0] = max_steer
            u[1] = max_speed
        
        
        return u
    


class dynLongVelInputs( controller.StaticController ) :
    """ 
    Simple proportionnal compensator
    ---------------------------------------
    r  : reference signal_proc vector  k x 1
    y  : sensor signal_proc vector     k x 1
    u  : control inputs vector    k x 1
    t  : time                     1 x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - y) * gain

    """
    
    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################
    
    
    ############################
    def __init__(self, k = 1):
        """ """
        
        # Dimensions
        self.k = 5   
        self.m = 2   
        self.p = 5
        
        controller.StaticController.__init__(self, self.k, self.m, self.p)
        
        # Label
        self.name = 'Proportionnal Controller'
        
        # Gains
        self.gain = 1
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        max_steer = 0.15
        max_speed = 15.0000
        u = np.zeros(self.m) # State derivative vector
        
        if t<1:
            u[0] = 0
            #u[0] = max_steer
            u[1] = max_speed
            #u[1] = 0
        elif(t>=1 and t<=3):
            u[0] = float(max_steer/2.0000*(t-1.0000))
            #u[0] = max_steer
            #u[1] = float(max_speed/40.0000*(t-10.0000))
            u[1] = max_speed
        elif(t>3 and t<=90):
            #u[0] = float(max_steer-max_steer/40.0000*(t-50.0000))
            u[0] = max_steer
            u[1] = max_speed
        else:
            #u[0] = 0
            u[0] = max_steer
            u[1] = max_speed
        
        
        return u

class dynLongForcesInputs( controller.StaticController ) :
    """ 
    Simple proportionnal compensator
    ---------------------------------------
    r  : reference signal_proc vector  k x 1
    y  : sensor signal_proc vector     k x 1
    u  : control inputs vector    k x 1
    t  : time                     1 x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - y) * gain

    """
    
    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################
    
    
    ############################
    def __init__(self, k = 1):
        """ """
        
        # Dimensions
        self.k = 6 
        self.m = 3   
        self.p = 6
        
        super().__init__(self.k, self.m, self.p)
        
        # Label
        self.name = 'Proportionnal Controller'
        
        # Gains
        self.gain = 1
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) # State derivative vector
        steer_max = 0.3
        
        if (t>=0 and t<25):
            u[0] = steer_max/25*t
            u[2] = 1500
        elif(t>=25 and t<50):
            u[0] = steer_max
            #u[0] = steer_max-steer_max/25*(t-25)
            u[2] = 0
        else:
            #u[0] = 0
            u[0] = steer_max
            u[2] = 0
        u[1] = 0
        
        
        return u
    
class fullDynTorqueInputs( controller.StaticController ) :
    """ 
    Simple proportionnal compensator
    ---------------------------------------
    r  : reference signal_proc vector  k x 1
    y  : sensor signal_proc vector     k x 1
    u  : control inputs vector    k x 1
    t  : time                     1 x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - y) * gain

    """
    
    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################
    
    
    ############################
    def __init__(self, k = 1):
        """ """
        
        # Dimensions
        self.k = 8  
        self.m = 3   
        self.p = 8
        
        super().__init__(self.k, self.m, self.p)
        
        # Label
        self.name = 'Proportionnal Controller'
        
        # Gains
        self.gain = 1
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) # State derivative vector
        #steer_max = 0.3
        torque_max = 1000
        if t<10:
            u[0] = 0
            u[2] = 1000
        elif(t>=10 and t<20):
            u[0] = 0
            #u[0] = steer_max/40*(t-10)
            u[2] = 1000-1000/10.000*(t-10)
        elif(t>=20 and t<25):
            u[0] = 0
            u[2] = 0
        elif(t>=25 and t<35):
            u[0] = 0
            #u[0] = steer_max
            #u[2] = 0
            u[2] = -torque_max/10.0000*(t-25)
        else:
            u[0] = 0
            u[2] = -torque_max
        u[1] = 0
        
        
        return u
    
class fullDynVoltInputs( controller.StaticController ) :
    """ 
    Simple proportionnal compensator
    ---------------------------------------
    r  : reference signal_proc vector  k x 1
    y  : sensor signal_proc vector     k x 1
    u  : control inputs vector    k x 1
    t  : time                     1 x 1
    ---------------------------------------
    u = c( y , r , t ) = (r - y) * gain

    """
    
    ###########################################################################
    # The two following functions needs to be implemented by child classes
    ###########################################################################
    
    
    ############################
    def __init__(self, k = 1):
        """ """
        
        # Dimensions
        self.k = 7  
        self.m = 2   
        self.p = 7
        
        super().__init__(self.k, self.m, self.p)
        
        # Label
        self.name = 'Proportionnal Controller'
        
        # Gains
        self.gain = 1
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) # State derivative vector
        #steer_max = 0.3
        volt_max = 8
        steer_max = 0.3
        if t<10:
            u[0] = 0
            #u[1] = 0
            u[1] = 5
        elif(t>=10 and t<20):
            u[0] = steer_max/10.000*(t-10)
            #u[1] = volt_max/10.000*(t-10)
            u[1] = 5
        elif(t>=20 and t<25):
            u[0] = steer_max
            #u[1] = volt_max
            u[1] = 5
        elif(t>=25 and t<35):
            u[0] = steer_max-steer_max/10.0000*(t-25)
            #u[1] = volt_max-volt_max/10.0000*(t-25)
            u[1] = 5
        else:
            u[0] = 0
            #u[1] = -volt_max
            u[1] = 5

        
        
        return u


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 13:41:26 2018

@author: Alexandre
"""

###############################################################################
from pyro.analysis import Trajectory
from pyro.control  import controller



###############################################################################
class OpenLoopController( controller.StaticController ) :
    """  
    Open-loop controller based on trajectory solution  
    ----------------------------------------------------
    u = c( x , r, ,t ) = u(t)
    
    """
    
    
    ############################
    def __init__(self, trajectory ):
        """ """
        
        # Sys
        self.trajectory = trajectory
        
        # Dimensions
        self.k = 1   
        self.m = trajectory.m
        self.n = trajectory.n

        if trajectory.y is not None:
            self.p = trajectory.y.shape[1]
        else:
            self.p = trajectory.x.shape[1]

        controller.StaticController.__init__( self, self.k, self.m, self.p)
        
        # Label
        self.name = 'Open Loop Controller'

    #############################
    def c( self , y , r , t  ):
        """  U depends only on time """
        
        u = self.trajectory.t2u( t )
        
        return u

    @staticmethod
    def load_from_file(name):
        traj = Trajectory.load(name)
        return OpenLoopController(traj)

    @property
    def time_final(self):
        return self.trajectory.time_final
    
    


###############################################################################
class Planner:
    """  
    Open-loop controller based on trajectory solution  
    ----------------------------------------------------
    u = c( x , r, ,t ) = u(t)
    
    """
    
    
    ############################
    def __init__(self, sys , cost_function = None ):
        """ """
        
        # Dynamic system model and constraints
        self.sys = sys
        
        # Cost function
        if cost_function is None:
            self.cost_function = sys.cost_function # default is quadratic cost
        else:
            self.cost_function = cost_function
        
        # Start and goal state
        self.x_start = sys.x0
        self.x_goal  = sys.xbar
        
        # Output variable
        self.traj = None
        
        
    ##############################
    def compute_solution(self):
        
        ################################################
        # Place holder: override this method in child classes
        raise NotImplementedError
        ################################################
        
        
    ##############################
    def show_solution(self, plot='xu', **kwargs):
        """ Plot computed trajectory solution """
        
        plotter = self.sys.get_plotter()
        
        return plotter.plot( self.traj, plot, **kwargs)
    
        
    ##############################
    def animate_solution(self, **kwargs):
        
        animator = self.sys.get_animator()
        self.ani = animator.animate_simulation( self.traj, **kwargs)
        
        return self.ani
    
        
    ##############################
    def animate_solution_to_html(self, **kwargs):
        
        animator = self.sys.get_animator()
        animator.animate_simulation( self.traj, show = False , **kwargs)
        
        return animator.ani.to_html5_video()
    
        
    ##############################
    def save_solution(self, name = 'planner_trajectory_solution.npy' ):
        
        self.traj.save( name )
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 25 12:16:59 2019

@author: alex
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import vehicle_steering
from pyro.control  import controller
###############################################################################

# Vehicule dynamical system
sys = vehicle_steering.ConstantSpeedKinematicCarModel()


class CarController( controller.StaticController ) :

    ############################
    def __init__(self):
        """ """
        
        controller.StaticController.__init__( self, 1, 1, 3)
        
        # Label
        self.name = 'Car Controller'
        
    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        """
        
        y_desired = r[0]
        
        lenght = 5.0
        speed = 2.0
        x_car = y[0]
        y_car = y[1]
        theta = y[2]
        
        
        # TODO your own feedback law below
        ########################
        delta = 1.5 * ( y_desired - y_car )
        delta = np.clip( delta , -1.5 , 1.5)
        ########################
        
        u = np.array([ delta ])
        
        return u
    

ctl = CarController()
    
cl_sys = ctl + sys




cl_sys.plot_phase_plane(1,2)
#cl_sys.plot_phase_plane_closed_loop(1,2)

cl_sys.x_ub = np.array([+8.,+8.,+8.])
cl_sys.x_lb = np.array([-8.,-8.,-8.])

# Animate the simulation
cl_sys.x0[1] = 0.5
cl_sys.x0[2] = 0.0
cl_sys.plot_trajectory()
anim = cl_sys.animate_simulation()
cl_sys.plot_phase_plane_trajectory(1,2)
#cl_sys.plot_phase_plane_trajectory_closed_loop(1,2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  1 19:51:49 2020

@author: alex
"""
import numpy as np
import matplotlib

from corom_robots import DrillingRobot
from corom_robots import DrillingRobotOnJig
from custom_drilling_controller      import CustomDrillingController


# Model dynamique du robot
sys = DrillingRobotOnJig()

# Controller
model = DrillingRobot()
ctl   = CustomDrillingController( model ) # Empty do nothing controller template

# Closed-loop dynamic
clsys = ctl + sys

# États initiaux
#clsys.x0 =  np.array([0.7,1.45,-1.4,0,0,0]) #  Tombe dans le trou
clsys.x0 =  np.array([0,1.4,-1.3,0,0,0]) #

# Simulation
tf = 6
n = 10001
clsys.compute_trajectory( tf , n)

clsys.plot_trajectory('x')
clsys.plot_trajectory('u')

# Données pour analyse
t        = clsys.traj.t
q_traj   = clsys.traj.x[:,0:3]  # Trajectoire des angles du robot
dq_traj  = clsys.traj.x[:,3:6]  # Trajectoire des vitesses du robot
tau_traj = clsys.traj.u         # Trajectoire des couples du robot

# Force de contact

f_traj   = np.zeros((n,3))
for i in range(n):
    f_traj[i,:] = sys.f_ext( q_traj[i,:] , dq_traj[i,:] )

fig , plots = matplotlib.pyplot.subplots(3)
plots[0].plot( t , f_traj[:,0] )
plots[1].plot( t , f_traj[:,1] )
plots[2].plot( t , f_traj[:,2] )
fig.canvas.manager.set_window_title('Contact forces')



clsys.animate_simulation( is_3d = True )




# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import manipulator
from pyro.control  import robotcontrollers
###############################################################################


# Model
torque_controlled_robot   = manipulator.ThreeLinkManipulator3D()


# Target
q_desired = np.array([0,0,0])

# joint PD 

model = torque_controlled_robot

joint_pd      = robotcontrollers.JointPD(3)
joint_pd.rbar = q_desired
joint_pd.kp   = np.array([100, 100 , 100])
joint_pd.kd   = np.array([100, 100 , 100])

# Closed-loops
robot_with_joint_pd = joint_pd + torque_controlled_robot 

# Simulation
robot_with_joint_pd.x0  = np.array([3.14,-3,2,0,0,0])
robot_with_joint_pd.compute_trajectory()
robot_with_joint_pd.plot_trajectory('x')
robot_with_joint_pd.plot_trajectory('u')
robot_with_joint_pd.animate_simulation(time_factor_video=1, is_3d=True)



# -*- coding: utf-8 -*-
"""
Created on Thu Oct 18 20:54:31 2018

@author: Alexandre
"""

import numpy as np

from pyro.dynamic import system


##############################################################################
        
class VanderPol( system.ContinuousDynamicSystem ):
    """ 
    SimpleIntegrator Example for a ContinuousDynamicSystem
    -------------------------------------------------------
    mass driven by a speed input: dx [m/sec] = f(x,u,t) = u [m/sec]
    
    """
    
    ############################
    def __init__(self, mu = 0.5 ):
        """ """
    
        # Dimensions
        self.n = 2   
        self.m = 1   
        self.p = 2
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__( self, self.n, self.m, self.p)
        
        # Labels
        self.name = 'Van der Pol oscillator'
        
        self.x_ub[0] = 3
        self.x_lb[0] = -3
        self.x_ub[1] = 3
        self.x_lb[1] = -3
        
        self.mu = mu
        
    
    #############################
    def f(self, x  , u  , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        y  = x[0]
        dy = x[1]
        
        
        ddy = - y + self.mu * dy * ( 1 - y**2 )
        
        dx = np.zeros(self.n) # State derivative vector
        
        dx[0] = dy
        dx[1] = ddy
        
        return dx


    
    
    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    
    sys = VanderPol()
    
    #sys.plot_phase_plane()
    
    sys.x0[1] = 4.0
    sys.compute_trajectory( tf = 20 )
    sys.plot_phase_plane_trajectory()
    #sys.plot_trajectory()
    
    sys.x0[1] = 0.1
    sys.compute_trajectory( tf = 20 )
    sys.plot_phase_plane_trajectory()


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:20:11 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import vehicle_steering
from pyro.planning import randomtree
###############################################################################

sys  = vehicle_steering.KinematicBicyleModel()

sys.dynamic_domain = False
###############################################################################

planner = randomtree.RRT( sys )

planner.x_start = np.array([0,0,0])
planner.x_goal  = np.array([0,1,0])

speed    = 2
steering = 0.2

planner.u_options = [
        np.array([ speed,-steering]),
        np.array([ speed,+steering]),
        np.array([ speed,0]),
        np.array([-speed,+steering]),
        np.array([-speed,0]),
        np.array([-speed,-steering])
        ]

planner.goal_radius       = 0.5
planner.dt                = 0.1
planner.steps             = 3
planner.max_solution_time = 8.0

planner.compute_solution()

planner.plot_tree()
planner.plot_tree_3d()
planner.show_solution()
planner.animate_solution()


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import pendulum
from pyro.planning import plan
from pyro.analysis import simulation
from pyro.control  import lqr
###############################################################################

sys  = pendulum.DoublePendulum()

traj = simulation.Trajectory.load('double_pendulum_directcollocation_hires.npy')

ctl = lqr.TrajectoryLQRController( sys , traj )

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0 = traj.x[0,:] + np.array([-0.2,0.1,0.1,-0.02])
cl_sys.compute_trajectory(10)
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 13 11:05:07 2018

@author: Alexandre
"""


###############################################################################
import numpy as np
###############################################################################
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer
from pyro.dynamic  import vehicle_steering
###############################################################################

sys  = vehicle_steering.KinematicCarModelwithObstacles()

# Set domain
sys.x_ub = np.array([+35, +3, +3])
sys.x_lb = np.array([-5, -2, -3])

sys.u_ub = np.array([+3, +1])
sys.u_lb = np.array([-3, -1])

# Discrete world
grid_sys = discretizer.GridDynamicSystem(sys, (51, 51, 21), (3, 3), 0.1)

# Cost Function
xbar    = np.array( [ 0, 0, 0] ) 
cf      = costfunction.TimeCostFunction( xbar )
cf.INF  = 50
cf.EPS  = 0.5

# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, cf )

dp.alpha = 1.0
dp.solve_bellman_equation( tol = 0.1 )

ctl = dp.get_lookup_table_controller()

ctl.plot_control_law( sys = sys , n = 100)

##############################################################################

# Simulation and animation
cl_sys = ctl + sys
cl_sys.x0   = np.array([30,0,0])
cl_sys.compute_trajectory( 10, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import pendulum
from pyro.control  import controller
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = pendulum.SinglePendulum()

sys.x_ub[0] = +4.0
sys.x_lb[0] = -4.0
sys.x_ub[1] = +5.0
sys.x_lb[1] = -5.0

#sys.u_ub[0] = +5.0
#sys.u_lb[0] = -5.0

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [41,21] , [3] , dt = 0.2 )

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys(sys)

qcf.xbar   = np.array([ -3.14 , 0 ]) # target
qcf.INF  = 300
qcf.EPS  = 1.0

# DP algo

dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)

dp.solve_bellman_equation( tol = 1.0 )

dp.plot_cost2go()
dp.plot_policy()


#asign controller
ctl    = dp.get_lookup_table_controller()
cl_sys = controller.ClosedLoopSystem( sys , ctl )

##############################################################################

# Simulation and animation
cl_sys.x0   = np.array([0,0])
cl_sys.compute_trajectory( 30, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import pendulum
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = pendulum.SinglePendulum()

sys.xbar = np.array([ -3.14 , 0 ]) 

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [101,101] , [3] )

# Cost Function
cf = costfunction.Reachability( sys.isavalidstate , sys.xbar )

# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, cf)

dp.solve_bellman_equation( animate_cost2go = True )



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 12:05:08 2018

@author: Alexandre
"""
###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import pendulum
from pyro.control import nonlinear
###############################################################################

sys  = pendulum.SinglePendulum()
ctl  = nonlinear.ComputedTorqueController( sys )

# Set Point
q_target = np.array([3.14])
ctl.rbar = q_target

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0  = np.array([-2,0])
cl_sys.compute_trajectory()
cl_sys.plot_phase_plane_trajectory()
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Wed May  8 08:47:05 2019

@author: alxgr
"""

###############################################################################
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
###############################################################################
from pyro.dynamic import system
from pyro.dynamic import mechanical
from pyro.dynamic import pendulum
###############################################################################


###############################################################################
        
class Manipulator( mechanical.MechanicalSystem ):
    """ 
    Manipulator Robot 
    -------------------------------------------------------
    
    Dynamics:
    H(q) ddq + C(q,dq) dq + d(q,dq) + g(q) = B(q) u + J(q)^T f_ext
    
    Foward kinematic end-effector:
    r = foward_kinematic_effector(q)
    
    Foward differential kinematic effector:
    dr = J(q) dq
    
    Dimensions
    -------------------------------------------------------
    dof : number of degrees-of-freedom of the system
    n   : number of dynamics states (2 * dof)
    e   : number of effector dof
    m   : number of actuators inputs     
    -------------------------------------------------------
    
    Vectors
    -------------------------------------------------------
    q      :  dim = (dof, 1)   : position variables 
    dq     :  dim = (dof, 1)   : velocity variables     
    ddq    :  dim = (dof, 1)   : acceleration variables
    r      :  dim = (e, 1)     : end-effector positions
    dr     :  dim = (e, 1)     : end-effector velocities
    u      :  dim = (m, 1)     : force input variables
    f_ext  :  dim = (e, 1)     : end-effector external forces
    d      :  dim = (dof, 1)   : state-dependent dissipative forces
    g      :  dim = (dof, 1)   : state-dependent conservatives forces
    -------------------------------------------------------
    
    Matrix
    -------------------------------------------------------
    H(q)   :  dim = (dof, dof) : inertia matrix
    C(q)   :  dim = (dof, dof) : corriolis matrix
    B(q)   :  dim = (dof, m)   : actuator matrix
    J(q)   :  dim = (e , dof)  : end-effector Jacobian matrix
    -------------------------------------------------------
    
    """
    
    ############################
    def __init__(self, dof = 1 , m = 1 , e = 1):
        """ """
        
        # Effector space dimensions
        self.e = e
               
        # initialize standard params
        mechanical.MechanicalSystem.__init__( self , dof )
        
        # Name
        self.name = str(dof) + 'Joint Manipulator Robot'
        
        # Default Label and units
        self.effector_label = []
        self.effector_units = []
        for i in range(e):
            self.effector_label.append('Axis '+str(i))
            self.effector_units.append('[m]')
        
    ###########################################################################
    # The following functions needs to be overloaded by child classes
    # to represent the dynamic of the system
    ###########################################################################
    
    # In Mother "Mechanical System" Class
    ###########################################################################
    # def H(self, q ):
    # def C(self, q , dq ):
    # def B(self, q ):
    # def g(self, q ):
    # def d(self, q , dq ):
    ###########################################################################
    
    # Specific to "Manipulator" Class
    
    ##############################
    def forward_kinematic_effector(self, q ):
        """ 
        Foward kinematic of the end-effector
        -------------------------------------
        r : effector position vector
        q : joint space vector
        """
        
        r = np.zeros( self.e ) # Place holder
        
        return r
    
    ##############################
    def J(self, q ):
        """
        Jacobian from joint space to effector space
        -------------------------------------------
        J : e x dof
        e : nb of effector position variables
        dof: number of configuration variables
        """
        
        J = np.zeros( ( self.e  , self.dof ) ) # Place holder
        
        return J
    
    ##############################
    def f_ext(self, q , dq , t = 0 ):
        """ 
        External force dependent on state or time
        
        """
        
        f_ext = np.zeros( self.e ) # Default is zero vector
        
        return f_ext
    
    ###########################################################################
    # No need to overwrite the following functions for custom system
    ###########################################################################
    
    ##############################
    def forward_differential_kinematic_effector(self, q, dq ):
        """ 
        End-effector foward differential kinematic
        ------------------------------------------
        dr : time derivative of effector position vector
        """
        
        dr = np.dot( self.J(q) , dq )
        
        return dr
        
    
    ##############################
    def generalized_forces(self, q  , dq  , ddq , t = 0 ):
        """ Computed generalized forces given a trajectory """  
        
        H = self.H( q )
        C = self.C( q , dq )
        g = self.g( q )
        d = self.d( q , dq )
        
        f_ext = self.f_ext( q , dq , t )
        J     = self.J( q )
                
        # Generalized forces
        forces = ( np.dot( H , ddq ) + np.dot( C , dq ) + g + d 
                   - np.dot( J.T , f_ext ) )
        
        return forces
    
    
    ##############################
    def actuator_forces(self, q  , dq  , ddq , t = 0 ):
        """ 
        Computed actuator forces given a trajectory (inverse dynamic) 
        -------------------------------------------------------------
        Note: will fail if B is not square (under-actuated system)
        """  
        
        # Actuator Matrix
        B = self.B( q )
                
        # Generalized forces
        forces = self.generalized_forces( q , dq , ddq , t )
        
        # Actuator forces
        u = np.dot( np.linalg.inv( B ) , forces )
        
        return u
    
    
    ##############################
    def ddq(self, q , dq , u , t = 0 ):
        """ 
        Computed accelerations given actuator forces (foward dynamic)
        --------------------------------------------------------------
        
        """  
        
        H = self.H( q )
        C = self.C( q , dq )
        g = self.g( q )
        d = self.d( q , dq)
        B = self.B( q )
        
        f_ext = self.f_ext( q , dq , t )
        J     = self.J( q )
        
        ddq = np.dot( np.linalg.inv( H ) ,  ( + np.dot( B   , u )
                                              + np.dot( J.T , f_ext )
                                              - np.dot( C   , dq ) 
                                              - g 
                                              - d ) )
        return ddq
    
    
    ##############################
    def plot_end_effector_trajectory(self, traj = None ):
        """ 
        Plot the end effector trajectory with respect to time
        --------------------------------------------------------------
        
        """  
        
        # If no argument is passed, use object traj
        if traj == None:
            traj = self.traj
        
        # Plot param
        fontsize = 5
        figsize  = (4, 3)
        dpi      = 300
        
        #Number of plot
        l = self.e * 1
        
        simfig , plots = plt.subplots( l, sharex=True, figsize=figsize,
                                      dpi=dpi, frameon=True)
            
        simfig.canvas.manager.set_window_title('End-Effector trajectory for ' + 
                                       self.name)
        
        
        # Computing end-effector trajectory
        n  = traj.time_steps
        
        r_traj  = np.zeros((n,self.e))
        dr_traj = np.zeros((n,self.e))
        
        for i in range(traj.time_steps):
            
            x = traj.x[i, :]
            
            q, dq = self.x2q(x)
            
            r  = self.forward_kinematic_effector( q )
            dr = self.forward_differential_kinematic_effector(q, dq)
            
            r_traj[i,:]  =  r
            dr_traj[i,:] = dr
        
        #j = 0 # plot index
        for i in range( l ):
                plots[i].plot( traj.t , r_traj[:,i] , 'b')
                plots[i].set_ylabel(self.effector_label[i] +'\n'+
                self.effector_units[i] , fontsize = fontsize )
                plots[i].grid(True)
                plots[i].tick_params( labelsize = fontsize )
        
        plots[l-1].set_xlabel('Time [sec]', fontsize=fontsize )
        
        simfig.tight_layout()
        simfig.canvas.draw()
        plt.show()
                
        return simfig
    
    
    ##############################
    def plot_manipulability_ellipsoid(self, q ):
        """ 
        Plot the manipulability ellispoid based on the jacobian
        --------------------------------------------------------------
        2D only
        
        """  
        
        # Plot robot config & base figure
        animator = self.get_animator()
        animator.show( q )
        fig = animator.showfig
        ax  = animator.showax
        
        # End-effector position
        r = self.forward_kinematic_effector( q )
        
        JJT     = self.J(q) @ self.J(q).T
        
        #u,s,v = np.linalg.svd( self.J(q) )
        l , v = np.linalg.eig( JJT )
        
        theta = np.arctan2( v[1,0] , v[0,0] ) * 180/3.1416
        w = np.sqrt(l[0])+0.01
        h = np.sqrt(l[1])+0.01
        e = matplotlib.patches.Ellipse(xy=(r[0], r[1]), width=w, height=h, angle=theta)
        e.set_clip_box(ax.bbox)
        e.set_alpha(0.1)
        e.set_facecolor('b')
        ax.add_artist(e)
        
        try:
            JJT_inv = np.linalg.inv(JJT)
            l , v   = np.linalg.eig( JJT_inv )
            
            #u,s,v = np.linalg.svd( B )
            
            theta = np.arctan2( v[1,0] , v[0,0] ) * 180/3.1416
            
            scale = 0.1 # Arbitrary number only for qualitative representation
            
            w =  np.sqrt(l[0])+0.01
            h =  np.sqrt(l[1])+0.01
            
            w_s = w * scale
            h_s = h * scale
            
            e = matplotlib.patches.Ellipse(xy=(r[0], r[1]), width=w_s, height=h_s, angle=theta)
            e.set_clip_box(ax.bbox)
            e.set_alpha(0.1)
            e.set_facecolor('r')
            ax.add_artist(e)
            
        except:
            print('Force ellispoid was not plotted since the configuration is singular')
        
        fig.canvas.draw_idle()
        plt.show()
        
        





###############################################################################
        
    
###############################################################################
        
class SpeedControlledManipulator( system.ContinuousDynamicSystem ):
    """ 
    Speed Controlled Manipulator Robot 
    -------------------------------------------------------
    
    This class can be used to model the high-level kinematic behavior of
    manipulator robot with good low-level velocity-controlled joint. All
    dynamic is neglected and inputs are considered as joint velocities.
    
    
    Dynamics:
    dq = u
    
    Foward kinematic end-effector:
    r = foward_kinematic_effector(q)
    
    Foward differential kinematic effector:
    dr = J(q) dq
    
    Dimensions
    -------------------------------------------------------
    dof : number of degrees-of-freedom of the system
    n   : number of dynamics states (2n = dof)
    e   : number of effector dof
    m   : number of actuators inputs     
    -------------------------------------------------------
    
    Vectors
    -------------------------------------------------------
    q = x  :  dim = (dof, 1)   : position variables 
    dq = u :  dim = (dof, 1)   : velocity variables     
    r      :  dim = (e, 1)     : end-effector positions
    dr     :  dim = (e, 1)     : end-effector velocities
    -------------------------------------------------------
    
    Matrix
    -------------------------------------------------------
    J(q)   :  dim = (e , dof)  : end-effector Jacobian matrix
    -------------------------------------------------------
    
    """
    ############################
    def __init__(self, dof , e ):
        """ """
        
        # Effector space dimensions
        self.e = e

        # Degree of Freedom
        self.dof = dof
        
        # Nb of states
        n = dof
        
        # initialize standard params
        system.ContinuousDynamicSystem.__init__( self, dof, dof, dof)
        
        # Name
        self.name = str(n) + ' Joint Speed Controlled Manipulator'
        
        # Labels, bounds and units
        for i in range(n):
            # joint angle states
            self.x_ub[i] = + np.pi * 2
            self.x_lb[i] = - np.pi * 2
            self.state_label[i] = 'Angle '+ str(i)
            self.state_units[i] = '[rad]'
        for i in range(n):
            # joint velocity inputs
            self.u_ub[i] = + np.pi * 2
            self.u_lb[i] = - np.pi * 2
            self.input_label[i] = 'Velocity ' + str(i)
            self.input_units[i] = '[rad/sec]'
            
    
    ############################
    @classmethod
    def from_manipulator(cls, Manipulator):
        """ From ContinuousDynamicSystem instance """
        
        instance = cls( Manipulator.dof , Manipulator.e )
        
        instance.forward_kinematic_lines      = Manipulator.forward_kinematic_lines
        instance.forward_kinematic_lines_plus = Manipulator.forward_kinematic_lines_plus
        instance.forward_kinematic_domain     = Manipulator.forward_kinematic_domain
        instance.forward_kinematic_effector   = Manipulator.forward_kinematic_effector
        instance.J                            = Manipulator.J
        instance.isavalidstate                = Manipulator.isavalidstate
        
        return instance
        
            
    ###########################################################################
    def f(self, x , u , t = 0 ):
        """ 
        Continuous time foward dynamics evaluation
        
        dx = f(x,u,t)
        
        INPUTS
        x  : state vector             n x 1
        u  : control inputs vector    m x 1
        t  : time                     1 x 1
        
        OUPUTS
        dx : state derivative vectror n x 1
        
        """
        
        dx = u       
        
        return dx
    
    ##############################
    def forward_differential_kinematic_effector(self, q, dq ):
        """ 
        End-effector foward differential kinematic
        ------------------------------------------
        dr : time derivative of effector position vector
        """
        
        dr = np.dot( self.J(q) , dq )
        
        return dr
    
    ##############################
    def forward_kinematic_effector(self, q ):
        """ 
        Foward kinematic of the end-effector
        -------------------------------------
        r : effector position vector
        q : joint space vector
        """
        
        r = np.zeros( self.e ) # Place holder
        
        return r
    
    ##############################
    def J(self, q ):
        """
        Jacobian from joint space to effector space
        -------------------------------------------
        J : e x dof
        e : nb of effector position variables
        dof: number of configuration variables
        """
        
        J = np.zeros( ( self.e  , self.dof ) ) # Place holder
        
        return J
            


###############################################################################
# One Link Manipulator
###############################################################################
        
class OneLinkManipulator( Manipulator ):
    """ 

    """
    
    ############################
    def __init__(self):
        """ """
        
        # Dimensions
        dof = 1
        m   = 1
        e   = 2
               
        # initialize standard params
        Manipulator.__init__( self, dof , m , e)
        
        # Name
        self.name = 'One Link Manipulator'
        
        # Plot param
        self.linestyle = '-'
                
        # Model parameters
        self.l1      = 2.5  # Link length
        self.lc1     = 1.2  # Center of mass distance from pivot
        self.m1      = 1    # Mass
        self.I1      = 0    # Inertia
        self.gravity = 9.81 # Gravity field constant
        self.d1      = 0.1  # Linear damping coef
        
        
    ##############################
    def trig(self, q ):
        """ 
        Compute cos and sin usefull in other computation 
        ------------------------------------------------
        """
        
        c1  = np.cos( q[0] )
        s1  = np.sin( q[0] )
        
        return [c1,s1]
    
    
    ##############################
    def forward_kinematic_effector(self, q ):
        
        [c1,s1] = self.trig( q )
        
        x = self.l1 * s1 
        y = self.l1 * c1 
        
        r = np.array([x,y])
        
        return r
    
    
    ##############################
    def J(self, q ):
        
        [c1,s1] = self.trig( q )
        
        J = np.zeros( ( self.e  , self.dof ) ) # Place holder
        
        J[0] = + self.l1 * c1 
        J[1] = - self.l1 * s1 
        
        return J
    
    
    ##############################
    def f_ext(self, q , dq , t = 0 ):
        
        f_ext = np.zeros( self.e ) # Default is zero vector
        
        return f_ext
        
    
    ###########################################################################
    def H(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * dq^T * H(q) * dq
        
        """  
        
        [c1,s1] = self.trig( q )
        
        H = np.zeros((1,1))
        
        H[0,0] = ( self.m1 * self.lc1**2 + self.I1  )
        
        return H
    
    
    ###########################################################################
    def C(self, q , dq ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        C = np.zeros((1,1))
        
        C[0,0] = 0

        return C
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.diag( np.ones( self.dof ) ) #  identity matrix
        
        return B
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        [c1,s1] = self.trig( q )
        
        g1 = (self.m1 * self.lc1  ) * self.gravity
        
        g = np.zeros(1)
        
        g[0] = - g1 * s1 

        return g
    
        
    ###########################################################################
    def d(self, q , dq ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        D = np.zeros((1,1))
        
        D[0,0] = self.d1
        
        d = np.dot( D , dq )
        
        return d
    
        
    ###########################################################################
    # Graphical output
    ###########################################################################
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = self.l1 * 1.2
        
        domain  = [ (-l,l) , (-l,l) , (-l,l) ]#  
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###############################
        # ground line
        ###############################
        
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([-10,0,0])
        pts[1,:] = np.array([+10,0,0])
        
        lines_pts.append( pts )
        lines_style.append('--')
        lines_color.append('k')
        
        ###########################
        # pendulum kinematic
        ###########################
        
        pts      = np.zeros(( 6 , 3 ))
        
        [c1,s1] = self.trig( q )
        
        l = self.l1 * 0.9
        h = self.l1 * 0.07
        
        pts[1,0] = 0 * s1 + h * c1
        pts[1,1] = 0 * c1 - h * s1
        
        pts[2,0] = l * s1 + h * c1
        pts[2,1] = l * c1 - h * s1
        
        pts[3,0] = l * s1 - h * c1
        pts[3,1] = l * c1 + h * s1
        
        pts[4,0] = 0 * s1 - h * c1
        pts[4,1] = 0 * c1 + h * s1
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
        
        
        ###########################
        # end effector
        ###########################
        
        pts      = np.zeros(( 7 , 3 ))
        
        pts[0,0] = l * s1 + 0 * c1
        pts[0,1] = l * c1 - 0 * s1
        
        pts[1,0] = (l+h) * s1 + 0 * c1
        pts[1,1] = (l+h) * c1 - 0 * s1
        
        pts[2,0] = (l+h) * s1 - h * c1
        pts[2,1] = (l+h) * c1 + h * s1
        
        pts[3,0] = (l+h+h) * s1 - h * c1
        pts[3,1] = (l+h+h) * c1 + h * s1
        
        pts[4,0] = (l+h) * s1 - h * c1
        pts[4,1] = (l+h) * c1 + h * s1
        
        pts[5,0] = (l+h) * s1 + h * c1
        pts[5,1] = (l+h) * c1 - h * s1
        
        pts[6,0] = (l+h+h) * s1 + h * c1
        pts[6,1] = (l+h+h) * c1 - h * s1
        
        
        lines_pts.append( pts )
        lines_style.append('-')
        lines_color.append('b')
                
        return lines_pts , lines_style , lines_color
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        
        x[0] = np.pi - x[0]
        u[0] = -u[0]
        

        return pendulum.SinglePendulum.forward_kinematic_lines_plus(self, x, u , t)





###############################################################################
# Two Link Manipulator
###############################################################################
        
class TwoLinkManipulator( Manipulator ):
    """  """
    
    ############################
    def __init__(self):
        """ """
        
        # Dimensions
        dof = 2
        m   = 2
        e   = 2
               
        # initialize standard params
        Manipulator.__init__( self, dof , m , e)
        
        # Name
        self.name = 'Two Link Manipulator'
        
        # params
        self.setparams()
        
        # Graphic output
        self.l_domain = 1.
                
            
    #############################
    def setparams(self):
        """ Set model parameters here """
        
        self.l1  = 0.5
        self.l2  = 0.3
        self.lc1 = 0.2
        self.lc2 = 0.1
        
        self.m1 = 1
        self.I1 = 0
        self.m2 = 1
        self.I2 = 0
        
        self.gravity = 9.81
        
        self.d1 = 0.5
        self.d2 = 0.5
        
        
    ##############################
    def trig(self, q ):
        """ 
        Compute cos and sin usefull in other computation 
        ------------------------------------------------
        """
        
        c1  = np.cos( q[0] )
        s1  = np.sin( q[0] )
        c2  = np.cos( q[1] )
        s2  = np.sin( q[1] )
        c12 = np.cos( q[0] + q[1] )
        s12 = np.sin( q[0] + q[1] )
        
        return [c1,s1,c2,s2,c12,s12]
    
    
    ##############################
    def forward_kinematic_effector(self, q ):
        """ """
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        x = self.l1 * s1 + self.l2 * s12 # x
        y = self.l1 * c1 + self.l2 * c12 # y
        
        r = np.array([x,y])
        
        return r
    
    
    ##############################
    def J(self, q ):
        """ """
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        J = np.zeros( ( self.e  , self.dof ) ) # Place holder
        
        J[0,0] = + self.l1 * c1 + self.l2 * c12
        J[0,1] =                  self.l2 * c12
        J[1,0] = - self.l1 * s1 - self.l2 * s12
        J[1,1] =                - self.l2 * s12
        
        return J
    
    
    ##############################
    def f_ext(self, q , dq , t = 0 ):
        """ """
        
        f_ext = np.zeros( self.e ) # Default is zero vector
        
        return f_ext
        
    
    ###########################################################################
    def H(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * dq^T * H(q) * dq
        
        """  
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        H = np.zeros((2,2))
        
        H[0,0] = ( self.m1 * self.lc1**2 + self.I1 + self.m2 * ( self.l1**2 
                 + self.lc2**2 + 2 * self.l1 * self.lc2 * c2 ) + self.I2 )
        H[1,0] = ( self.m2 * self.lc2**2 + self.m2 * self.l1 * self.lc2 * c2 
                   + self.I2 )
        H[0,1] = H[1,0]
        H[1,1] = self.m2 * self.lc2 ** 2 + self.I2
        
        return H
    
    
    ###########################################################################
    def C(self, q , dq ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        h = self.m2 * self.l1 * self.lc2 * s2
        
        C = np.zeros((2,2))
        
        C[0,0] = - h  * dq[1]
        C[1,0] =   h  * dq[0]
        C[0,1] = - h * ( dq[0] + dq[1] )
        C[1,1] = 0

        return C
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.diag( np.ones( self.dof ) ) #  identity matrix
        
        return B
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        g1 = (self.m1 * self.lc1 + self.m2 * self.l1 ) * self.gravity
        g2 = self.m2 * self.lc2 * self.gravity
        
        G = np.zeros(2)
        
        G[0] = - g1 * s1 - g2 * s12
        G[1] = - g2 * s12

        return G
    
        
    ###########################################################################
    def d(self, q , dq ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        D = np.zeros((2,2))
        
        D[0,0] = self.d1
        D[1,0] = 0
        D[0,1] = 0
        D[1,1] = self.d2
        
        d = np.dot( D , dq )
        
        return d
    
        
    ###########################################################################
    # Graphical output
    ###########################################################################
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = self.l_domain
        
        domain  = [ (-l,l) , (-l,l) , (-l,l) ]#  
                
        return domain
    
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###############################
        # ground line
        ###############################
        
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([-10,0,0])
        pts[1,:] = np.array([+10,0,0])
        
        lines_pts.append( pts )
        lines_style.append('--')
        lines_color.append('k')
        
        ###########################
        # pendulum kinematic
        ###########################
        
        pts      = np.zeros(( 3 , 3 ))
        pts[0,:] = np.array([0,0,0])
        
        [c1,s1,c2,s2,c12,s12] = self.trig( q )
        
        pts[1,0] = self.l1 * s1
        pts[1,1] = self.l1 * c1
        
        pts[2,0] = self.l1 * s1 + self.l2 * s12
        pts[2,1] = self.l1 * c1 + self.l2 * c12
        
        lines_pts.append( pts )
        lines_style.append('o-')
        lines_color.append('b')
                
        return lines_pts , lines_style , lines_color
    
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        
        #x[0] = x[0]
        #x[1] = np.pi - x[0]
        #u[0] = -u[0]
        

        return pendulum.DoublePendulum.forward_kinematic_lines_plus(self, x, u , t)
    


###############################################################################
# Three Link Manipulator
###############################################################################
        
class ThreeLinkManipulator3D( Manipulator ):
    """ 
    Three link Manipulator Class 
    -------------------------------
    
    base:     revolute arround z
    shoulder: revolute arround y
    elbow:    revolute arround y
    
    see Example 4.3 in
    http://www.cds.caltech.edu/~murray/books/MLS/pdf/mls94-manipdyn_v1_2.pdf
    """
    
    ############################
    def __init__(self):
        """ """
        
        # Dimensions
        dof = 3
        m   = 3
        e   = 3
               
        # initialize standard params
        Manipulator.__init__( self, dof , m , e)
        
        # Name
        self.name = 'Three Link Manipulator'
        
        # params
        self.setparams()
                
            
    #############################
    def setparams(self):
        """ Set model parameters here """
        
        # Kinematic
        self.l1  = 1 
        self.l2  = 1
        self.l3  = 1
        self.lc1 = 1
        self.lc2 = 1
        self.lc3 = 1
        
        # Inertia
        self.m1 = 1
        self.m2 = 1
        self.m3 = 1
        
        self.I1z = 1
        
        self.I2x = 1
        self.I2y = 1
        self.I2z = 1
        
        self.I3x = 1
        self.I3y = 1
        self.I3z = 1
        
        # Gravity
        self.gravity = 9.81
        
        # Joint damping
        self.d1 = 1
        self.d2 = 1
        self.d3 = 1
        
        # Total length
        self.lw  = (self.l1+self.l2+self.l3)
        
        
    ##############################
    def trig(self, q ):
        """ 
        Compute cos and sin usefull in other computation 
        ------------------------------------------------
        """
        
        c1  = np.cos( q[0] )
        s1  = np.sin( q[0] )
        c2  = np.cos( q[1] )
        s2  = np.sin( q[1] )
        c3  = np.cos( q[2] )
        s3  = np.sin( q[2] )
        c12 = np.cos( q[0] + q[1] )
        s12 = np.sin( q[0] + q[1] )
        c23 = np.cos( q[2] + q[1] )
        s23 = np.sin( q[2] + q[1] )
        
        return [c1,s1,c2,s2,c3,s3,c12,s12,c23,s23]
    
    
    ##############################
    def forward_kinematic_effector(self, q ):
        """ """
        
        [c1,s1,c2,s2,c3,s3,c12,s12,c23,s23] = self.trig( q )
        
        # Three robot points
        
        # Base of the robot
        p0 = [0,0,0]
        
        # Shperical point 
        p1 = [ 0, 0, self.l1 ]
        
        # Elbow
        z2 = self.l1 - self.l2 * s2
        
        r2 = self.l2 * c2
        x2 = r2 * c1
        y2 = r2 * s1
        
        p2 = [ x2, y2, z2 ]
        
        # End-effector
        z3 = self.l1 - self.l2 * s2 - self.l3 * s23
        
        r3 = self.l2 * c2 + self.l3 * c23
        x3 = r3 * c1
        y3 = r3 * s1
                
        r = np.array([x3, y3, z3])
        
        return r
    
    
    ##############################
    def J(self, q ):
        """ """
        
        [c1,s1,c2,s2,c3,s3,c12,s12,c23,s23] = self.trig( q )
        
        J = np.zeros((3,3))
        
        J[0,0] =  -( self.l2 * c2 + self.l3 * c23 ) * s1
        J[0,1] =  -( self.l2 * s2 + self.l3 * s23 ) * c1
        J[0,2] =  - self.l3 * s23 * c1
        J[1,0] =   ( self.l2 * c2 + self.l3 * c23 ) * c1
        J[1,1] =  -( self.l2 * s2 + self.l3 * s23 ) * s1
        J[1,2] =  - self.l3 * s23 * s1
        J[2,0] =  0
        J[2,1] =  -( self.l2 * c2 + self.l3 * c23 )
        J[2,2] =  - self.l3 * c23
        
        return J
    
    
    ##############################
    def f_ext(self, q , dq , t = 0 ):
        """ """
        
        f_ext = np.zeros( self.e ) # Default is zero vector
        
        return f_ext
        
    
    ###########################################################################
    def H(self, q ):
        """ 
        Inertia matrix 
        ----------------------------------
        dim( H ) = ( dof , dof )
        
        such that --> Kinetic Energy = 0.5 * dq^T * H(q) * dq
        
        """  
        
        [c1,s1,c2,s2,c3,s3,c12,s12,c23,s23] = self.trig( q )
        
        # variable to match the book notation
        
        m1 = self.m1
        m2 = self.m2
        m3 = self.m3
        
        Iz1 = self.I1z
        Ix2 = self.I2x
        Iy2 = self.I2y
        Iz2 = self.I2z
        Ix3 = self.I3x
        Iy3 = self.I3y
        Iz3 = self.I3z
        
        l1 = self.l2
        r1 = self.lc2
        l2 = self.l3
        r2 = self.lc3
        
        
        H = np.zeros((3,3))
        
        H[0,0] = (Iy2 * s2 **2 + Iy3 * s23 **2 + Iz1 
                 + Iz2 * c2 **2 + Iz3 * c23 **2 + m2 * ( r1 * c2 ) **2 
                 + m3 * ( l1 * c2 + r2 * c23 ) **2 )
        H[0,1] = 0
        H[0,2] = 0
        H[1,0] = 0
        H[1,1] = (Ix2 + Ix3 + m3 * l1 **2 + m2 * r1 **2 
                 + m3 * r2 **2 + 2 * m3 *l1 * r2 * c3)
        H[1,2] = Ix3 + m3 * r2 **2 + m3 * l1 * r2 * c3
        H[2,0] = 0
        H[2,1] = H[1,2]
        H[2,2] = Ix3 + m3 * r2 ** 2
        
        return H
    
    
    ###########################################################################
    def C(self, q , dq ):
        """ 
        Corriolis and Centrifugal Matrix 
        ------------------------------------
        dim( C ) = ( dof , dof )
        
        such that --> d H / dt =  C + C^T
        
        
        """ 
        
        [c1,s1,c2,s2,c3,s3,c12,s12,c23,s23] = self.trig( q )
        
        # variable to match the book notation
        
        m1 = self.m1
        m2 = self.m2
        m3 = self.m3
        
        Iz1 = self.I1z
        Ix2 = self.I2x
        Iy2 = self.I2y
        Iz2 = self.I2z
        Ix3 = self.I3x
        Iy3 = self.I3y
        Iz3 = self.I3z
        
        l1 = self.l2
        r1 = self.lc2
        l2 = self.l3
        r2 = self.lc3
        
        
        T112 = (( Iy2 - Iz2 - m2 * r1 **2 ) * c2 * s2 
                 + ( Iy3 - Iz3 ) * c23 * s23
                 - m3 * ( l1 * c2 + r2 * c23 ) * ( l1 * s2 + r2 * s23 ) )
        T113 = (( Iy3 - Iz3 ) * c23 * s23 
               - m3 * r2 * s23 * ( l1 * c2 + r2 * c23 ))
        T121 = T112
        T131 = T113
        T211 = (( Iz2 - Iy2 + m2 * r1 **2 ) * c2 * s2 
               + ( Iz3 - Iy3 ) * c23 * s23 + m3 * 
                ( l1 * c2 + r2 * c23 ) * ( l1 * s2 + r2 * s23 ))
        T223 = - l1 * m3 * r2 * s3
        T232 = T223
        T233 = T223
        T311 = (( Iz3 - Iy3 ) * c23 * s23 
                + m3 * r2 * s23 * ( l1 * c2 + r2 * c23 ))
        T322 = l1 * m3 * r2 * s3
                
        C = np.zeros((3,3))
        
        C[0,0] = T112 * dq[1] + T113 * dq[2]
        C[0,1] = T121 * dq[0]
        C[0,2] = T131 * dq[0]
        
        C[1,0] = T211 * dq[0]
        C[1,1] = T223 * dq[2]
        C[1,2] = T232 * dq[1] + T233 * dq[2]
        
        C[2,0] = T311 * dq[0]
        C[2,1] = T322 * dq[1]
        C[2,2] = 0 
        
        return C
    
    
    ###########################################################################
    def B(self, q ):
        """ 
        Actuator Matrix  : dof x m
        """
        
        B = np.diag( np.ones( self.dof ) ) #  identity matrix
        
        return B
    
    
    ###########################################################################
    def g(self, q ):
        """ 
        Gravitationnal forces vector : dof x 1
        """
        
        [c1,s1,c2,s2,c3,s3,c12,s12,c23,s23] = self.trig( q )
        
        G = np.zeros(3)
        
        g = self.gravity
        
        G[0] = 0
        G[1] = -(self.m2 * g * self.lc2 + self.m3 * g * self.l2 ) * c2 - self.m3 * g * self.lc3 * c23
        G[2] = - self.m3 * g * self.lc3 * c23

        return G
    
        
    ###########################################################################
    def d(self, q , dq ):
        """ 
        State-dependent dissipative forces : dof x 1
        """
        
        D = np.zeros((3,3))
        
        D[0,0] = self.d1
        D[1,1] = self.d2
        D[2,2] = self.d3
        
        d = np.dot( D , dq )
        
        return d
    
        
    ###########################################################################
    # Graphical output
    ###########################################################################
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = 2
        
        domain  = [ (-l,l) , (-l,l) , (0,l*2) ]#  
                
        return domain
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###############################
        # ground line
        ###############################
        
        pts      = np.zeros(( 5 , 3 ))
        pts[0,:] = np.array([-1,-1,0])
        pts[1,:] = np.array([+1,-1,0])
        pts[2,:] = np.array([+1,+1,0])
        pts[3,:] = np.array([-1,+1,0])
        pts[4,:] = np.array([-1,-1,0])
        
        lines_pts.append( pts )
        lines_style.append('--')
        lines_color.append('k')
        
        ###########################
        # robot kinematic
        ###########################
        
        pts      = np.zeros(( 4 , 3 ))
        pts[0,:] = np.array([0,0,0])
        
        [c1,s1,c2,s2,c3,s3,c12,s12,c23,s23] = self.trig( q )
        
        # Three robot points

        # Shperical point 
        pts[1,0] = 0
        pts[1,1] = 0
        pts[1,2] = self.l1 
        
        # Elbow
        z2 = self.l1 - self.l2 * s2
        
        r2 = self.l2 * c2
        x2 = r2 * c1
        y2 = r2 * s1
        
        pts[2,0] = x2
        pts[2,1] = y2
        pts[2,2] = z2
        
        # End-effector
        z3 = self.l1 - self.l2 * s2 - self.l3 * s23
        
        r3 = self.l2 * c2 + self.l3 * c23
        x3 = r3 * c1
        y3 = r3 * s1
                
        pts[3,0] = x3
        pts[3,1] = y3
        pts[3,2] = z3 

        lines_pts.append( pts )
        lines_style.append('o-')
        lines_color.append('b')
 
        return lines_pts , lines_style , lines_color
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        
        return [],[],[]



###############################################################################
# Five Planar Link Manipulator
###############################################################################
        
class FiveLinkPlanarManipulator( Manipulator ):
    """ 
    Note: Only kinematic functions 
    """
    
    ############################
    def __init__(self):
        """ """
        
        # Dimensions
        dof = 5
        m   = 5
        e   = 2
               
        # initialize standard params
        Manipulator.__init__( self, dof , m , e)
        
        # Name
        self.name = 'Five Link Manipulator'
        
        # params
        self.setparams()

        self.lines_plus = False
                
            
    #############################
    def setparams(self):
        """ Set model parameters here """
        
        l1  = 0.5
        l2  = 0.5
        l3  = 0.5
        l4  = 0.5
        l5  = 0.5
        
        self.l = np.array([l1,l2,l3,l4,l5])
        
        
    ##############################
    def trig(self, q ):
        """ 
        Compute cos and sin usefull in other computation 
        ------------------------------------------------
        """
        
        c1  = np.cos( q[0] )
        s1  = np.sin( q[0] )
        c2  = np.cos( q[1] )
        s2  = np.sin( q[1] )
        c3  = np.cos( q[2] )
        s3  = np.sin( q[2] )
        c4  = np.cos( q[3] )
        s4  = np.sin( q[3] )
        c5  = np.cos( q[4] )
        s5  = np.sin( q[4] )
        
        cos_rel = np.array( [ c1 , c2 , c3 , c4 , c5 ])
        sin_rel = np.array( [ s1 , s2 , s3 , s4 , s5 ])
        
        c12    = np.cos( q[0] + q[1] )
        s12    = np.sin( q[0] + q[1] )
        c123   = np.cos( q[0] + q[1] + q[2])
        s123   = np.sin( q[0] + q[1] + q[2])
        c1234  = np.cos( q[0] + q[1] + q[2] + q[3])
        s1234  = np.sin( q[0] + q[1] + q[2] + q[3])
        c12345 = np.cos( q[0] + q[1] + q[2] + q[3] + q[4])
        s12345 = np.sin( q[0] + q[1] + q[2] + q[3] + q[4])
        
        cos_abs = np.array( [ c1 , c12 , c123 , c1234 , c12345 ])
        sin_abs = np.array( [ s1 , s12 , s123 , s1234 , s12345 ])
        
        return [cos_rel,sin_rel,cos_abs,sin_abs]
    
    
    ##############################
    def forward_kinematic_effector(self, q ):
        """ """
        
        [cos_rel,sin_rel,cos_abs,sin_abs] = self.trig( q )
        
        x = (self.l * sin_abs).sum()
        y = (self.l * cos_abs).sum()
        
        r = np.array([x,y])
        
        return r
    
    
    ##############################
    def J(self, q ):
        """ """
        
        [cos_rel,sin_rel,cos_abs,sin_abs] = self.trig( q )
        
        J = np.zeros( ( self.e  , self.dof ) ) # Place holder
        
        J[0,0] = self.l[4] * cos_abs[4] + self.l[3] * cos_abs[3] + self.l[2] * cos_abs[2] + self.l[1] * cos_abs[1]  + self.l[0] * cos_abs[0] 
        J[0,1] = self.l[4] * cos_abs[4] + self.l[3] * cos_abs[3] + self.l[2] * cos_abs[2] + self.l[1] * cos_abs[1]             
        J[0,2] = self.l[4] * cos_abs[4] + self.l[3] * cos_abs[3] + self.l[2] * cos_abs[2]           
        J[0,3] = self.l[4] * cos_abs[4] + self.l[3] * cos_abs[3]                
        J[0,4] = self.l[4] * cos_abs[4] 
        
        J[1,0] = - self.l[4] * sin_abs[4] - self.l[3] * sin_abs[3] - self.l[2] * sin_abs[2] - self.l[1] * sin_abs[1]  - self.l[0] * sin_abs[0] 
        J[1,1] = - self.l[4] * sin_abs[4] - self.l[3] * sin_abs[3] - self.l[2] * sin_abs[2] - self.l[1] * sin_abs[1]             
        J[1,2] = - self.l[4] * sin_abs[4] - self.l[3] * sin_abs[3] - self.l[2] * sin_abs[2]           
        J[1,3] = - self.l[4] * sin_abs[4] - self.l[3] * sin_abs[3]                
        J[1,4] = - self.l[4] * sin_abs[4] 
        
        return J
    
        
    ###########################################################################
    # Graphical output
    ###########################################################################
    
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = 3
        
        domain  = [ (-l,l) , (-l,l) , (-l,l) ]#  
                
        return domain
    
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts   = [] # list of array (n_pts x 3) for each lines
        lines_style = []
        lines_color = []
        
        ###############################
        # ground line
        ###############################
        
        pts      = np.zeros(( 2 , 3 ))
        pts[0,:] = np.array([-10,0,0])
        pts[1,:] = np.array([+10,0,0])
        
        lines_pts.append( pts )
        lines_style.append('--')
        lines_color.append('k')
        
        ###########################
        # robot kinematic
        ###########################
        
        pts      = np.zeros(( 6 , 3 ))
        pts[0,:] = np.array([0,0,0])
        
        [cos_rel,sin_rel,cos_abs,sin_abs] = self.trig( q )
        
        pts[1,0] = self.l[0] * sin_abs[0]
        pts[2,0] = self.l[1] * sin_abs[1] + pts[1,0]
        pts[3,0] = self.l[2] * sin_abs[2] + pts[2,0]
        pts[4,0] = self.l[3] * sin_abs[3] + pts[3,0]
        pts[5,0] = self.l[4] * sin_abs[4] + pts[4,0]
        
        pts[1,1] = self.l[0] * cos_abs[0]
        pts[2,1] = self.l[1] * cos_abs[1] + pts[1,1]
        pts[3,1] = self.l[2] * cos_abs[2] + pts[2,1]
        pts[4,1] = self.l[3] * cos_abs[3] + pts[3,1]
        pts[5,1] = self.l[4] * cos_abs[4] + pts[4,1]
        
        lines_pts.append( pts )
        lines_style.append('o-')
        lines_color.append('b')
 
        return lines_pts , lines_style , lines_color
    
    ###########################################################################
    def forward_kinematic_lines_plus(self, x , u , t ):
        
        return [],[],[]


##############################################################################
# Two Planar Link Manipulator with obstacles
###############################################################################
        
class TwoLinkManipulatorwithObstacles( TwoLinkManipulator ):
    """
    Maniplator with non-allowable states based on end-effector position only
    """
    
    ############################
    def __init__(self):
        """ """
        # initialize standard params
        TwoLinkManipulator.__init__( self )
        
        self.l1 = 1.1
        self.l2 = 0.9
        
        # Labels
        self.name = 'Two Link Planar Manipulator with Obstacles'

        self.obstacles = [
                [ (-1, -1),( 3, -0.2)],
                [ (0.5, 0.5),(2, 3)],
                [ (-2, 0),(-0.2, 3)]
                ]
        
        
    #############################
    def isavalidstate(self , x ):
        """ check if x is in the state domain """
        ans = False
        for i in range( x.size ):
            ans = ans or ( x[i] < self.x_lb[i] )
            ans = ans or ( x[i] > self.x_ub[i] )
        
        # effector position
        q = self.xut2q( x, self.ubar, 0)
        r = self.forward_kinematic_effector( q )

        buffer = 0.0
        
        for obs in self.obstacles:
            on_obs = (( r[0] + buffer > obs[0][0]) and  
                      ( r[1] + buffer > obs[0][1]) and 
                      ( r[0] - buffer < obs[1][0]) and 
                      ( r[1] - buffer < obs[1][1]) )
                     
            ans = ans or on_obs
            
        return not(ans)
    
        
    ###########################################################################
    def forward_kinematic_domain(self, q ):
        """ 
        """
        l = 3
        
        domain  = [ (-l,l) , (-l,l) , (-l,l) ]#  
                
        return domain
        
       
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts , lines_style , lines_color = TwoLinkManipulator.forward_kinematic_lines(self, q )
        

        ###########################
        # obstacles
        ###########################
        
        for obs in self.obstacles:
            
            pts = np.zeros((5,3))
            
            pts[0,0] = obs[0][0]
            pts[0,1] = obs[0][1]
            
            pts[1,0] = obs[0][0]
            pts[1,1] = obs[1][1]
            
            pts[2,0] = obs[1][0]
            pts[2,1] = obs[1][1]
            
            pts[3,0] = obs[1][0]
            pts[3,1] = obs[0][1]
            
            pts[4,0] = obs[0][0]
            pts[4,1] = obs[0][1]
            
            lines_pts.append( pts )
            lines_style.append('-')
            lines_color.append('k')
            
                
        return lines_pts , lines_style , lines_color


###############################################################################
# Five Planar Link Manipulator with obstacles
###############################################################################
        
class FiveLinkPlanarManipulatorwithObstacles( FiveLinkPlanarManipulator ):
    """
    Maniplator with non-allowable states based on end-effector position only
    """
    
    ############################
    def __init__(self):
        """ """
        # initialize standard params
        FiveLinkPlanarManipulator.__init__( self )
        
        # Labels
        self.name = 'Five Link Planar Manipulator with Obstacles'

        self.obstacles = [
                [ (-1, -1),( 3, -0.2)],
                [ (0.5, 0.5),(2, 3)],
                [ (-2, 0),(-0.2, 3)]
                ]
        
        
    #############################
    def isavalidstate(self , x ):
        """ check if x is in the state domain """
        ans = False
        for i in range( x.size ):
            ans = ans or ( x[i] < self.x_lb[i] )
            ans = ans or ( x[i] > self.x_ub[i] )
        
        # effector position
        q = self.xut2q( x, self.ubar, 0)
        r = self.forward_kinematic_effector( q )

        buffer = 0.0
        
        for obs in self.obstacles:
            on_obs = (( r[0] + buffer > obs[0][0]) and  
                      ( r[1] + buffer > obs[0][1]) and 
                      ( r[0] - buffer < obs[1][0]) and 
                      ( r[1] - buffer < obs[1][1]) )
                     
            ans = ans or on_obs
            
        return not(ans)
        
       
    ###########################################################################
    def forward_kinematic_lines(self, q ):
        """ 
        Compute points p = [x;y;z] positions given config q 
        ----------------------------------------------------
        - points of interest for ploting
        
        Outpus:
        lines_pts = [] : a list of array (n_pts x 3) for each lines
        
        """
        
        lines_pts , lines_style , lines_color = FiveLinkPlanarManipulator.forward_kinematic_lines(self, q )
        

        ###########################
        # obstacles
        ###########################
        
        for obs in self.obstacles:
            
            pts = np.zeros((5,3))
            
            pts[0,0] = obs[0][0]
            pts[0,1] = obs[0][1]
            
            pts[1,0] = obs[0][0]
            pts[1,1] = obs[1][1]
            
            pts[2,0] = obs[1][0]
            pts[2,1] = obs[1][1]
            
            pts[3,0] = obs[1][0]
            pts[3,1] = obs[0][1]
            
            pts[4,0] = obs[0][0]
            pts[4,1] = obs[0][1]
            
            lines_pts.append( pts )
            lines_style.append('-')
            lines_color.append('k')
                
        return lines_pts , lines_style , lines_color
    
    
    
    
'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":     
    
    #sys = OneLinkManipulator()
    sys = TwoLinkManipulator()
    
    sys.x0[0]   = 0.1
    sys.ubar[0] = 4
    sys.ubar[1] = 4

    sys.plot_trajectory()
    sys.animate_simulation()
    sys.plot_end_effector_trajectory()
    
    
    #Ellispoid validation
    sys.plot_manipulability_ellipsoid( [0.0,0.0] )
    sys.plot_manipulability_ellipsoid( [1.0,0.1] )
    sys.plot_manipulability_ellipsoid( [2.0,0.8] )
    sys.plot_manipulability_ellipsoid( [3.0,1.6] )
    
    sys.l1 = np.sqrt(2)/2
    sys.l2 = 0.5
    sys.plot_manipulability_ellipsoid( [-np.pi/4, +3*np.pi/4] ) #equal
    
    
    #sys = ThreeLinkManipulator3D()
    #sys.x0[0] = 0.1
    #sys.animate_simulation( is_3d = True )
    #sys.plot_trajectory()
    
    #sys = FiveLinkPlanarManipulator()
    #sys.ubar = np.array([1,1,1,1,1])
    #sys.animate_simulation()
    
    #aaa = TwoLinkManipulatorwithObstacles()
    #bbb = SpeedControlledManipulator.from_manipulator( aaa )
    #bbb.ubar = np.array([1,1])
    #bbb.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic  import cartpole
from pyro.control  import nonlinear
from pyro.planning import randomtree
###############################################################################


sys  = cartpole.RotatingCartPole()

###############################################################################

x_start = np.array([0,-3.14,0,0])
x_goal  = np.array([0,0,0,0])

planner = randomtree.RRT( sys , x_start )

t = 50
    
planner.u_options = [
        np.array([-t,0]),
        np.array([ 0,0]),
        np.array([+t,0])
        ]

planner.goal_radius          = 1.5
planner.dt                   = 0.05
planner.max_nodes            = 5000
planner.max_solution_time    = 2.0
planner.max_distance_compute = 1000

planner.dyna_plot            = False

planner.find_path_to_goal( x_goal )

planner.plot_tree()

# Controller

ctl  = nonlinear.ComputedTorqueController( sys , planner.traj )

ctl.w0   = 1.0
ctl.zeta = 0.7

# goal
ctl.rbar = np.array([0,0])


# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0 = x_start
cl_sys.compute_trajectory(15)
cl_sys.plot_phase_plane_trajectory()
cl_sys.plot_trajectory( 'xu')
cl_sys.animate_simulation( time_factor_video=1.2, is_3d=True)


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.planning import randomtree
from pyro.dynamic  import manipulator
from pyro.analysis import simulation
from pyro.analysis import graphical
###############################################################################

torque_controlled_robot = manipulator.FiveLinkPlanarManipulatorwithObstacles()
speed_controlled_robot  = manipulator.SpeedControlledManipulator.from_manipulator( torque_controlled_robot )


traj = simulation.Trajectory.load( 'fivelinkplan.npy' )

animator = graphical.Animator( speed_controlled_robot )

animator.animate_simulation( traj )



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 12 20:28:17 2018

@author: Alexandre
"""
##############################################################################
import numpy as np
##############################################################################
from pyro.dynamic  import pendulum
from pyro.planning import discretizer
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming
##############################################################################


sys  = pendulum.SinglePendulum()

sys.x_ub = np.array([-3.14+6, +6])
sys.x_lb = np.array([-3.14-6, -6])

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [201,201] , [3] )

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys )

# Cost Function

xbar = np.array([-3.14,0])
tcf = costfunction.TimeCostFunction( xbar )
tcf.INF = 5.0
tcf.EPS = 0.1

sys.cost_function = tcf

# DP algo
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, tcf )

dp.solve_bellman_equation( tol = 0.05 , animate_cost2go = False )


dp.clean_infeasible_set()
dp.plot_cost2go_3D()
dp.plot_policy()

#dp.animate_policy( show = True , save = False )
#dp.animate_cost2go( show = False , save = True )
dp.animate_policy( show = False , save = True )


ctl = dp.get_lookup_table_controller()

#ctl.plot_control_law( sys = sys , n = 100)


#asign controller
cl_sys = ctl + sys
cl_sys.x0   = np.array([0., 0.])
cl_sys.compute_trajectory( 10, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import vehicle_steering
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = vehicle_steering.HolonomicMobileRobotwithObstacles()

#sys.obstacles[1][0] = (5,5)

#sys.x_ub[1] = 15
#sys.x_lb[1] = 0

# Discrete world 
grid_sys = discretizer.GridDynamicSystem( sys , [51,51] , [3,3] )

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys( sys )

qcf.xbar = np.array([ 10. , 0. ]) # target
qcf.Q[0,0] = 1.0
qcf.Q[1,1] = 1.0
qcf.R[0,0] = 0.0
qcf.R[1,1] = 0.0
qcf.S[0,0] = 10.0
qcf.S[1,1] = 10.0
qcf.INF    = 8000

# DP algo
#dp = dprog.DynamicProgramming( grid_sys, qcf )
dp = dynamicprogramming.DynamicProgrammingWithLookUpTable( grid_sys, qcf)

#dp.plot_cost2go()
#dp.solve_bellman_equation( tol = 0.01 )
dp.solve_bellman_equation( tol = 0.01 , animate_cost2go = True )
#dp.solve_bellman_equation( tol = 0.01 , animate_policy = True )

#grid_sys.plot_grid_value( dp.J_next )

ctl = dp.get_lookup_table_controller()

#ctl.plot_control_law( k=0 , sys = sys , n = 50)
#ctl.plot_control_law( k=1 , sys = sys , n = 50)


##############################################################################

# Simulation and animation
cl_sys      = ctl + sys
cl_sys.x0   = np.array([-8,5])
cl_sys.compute_trajectory( 60, 10001, 'euler')
cl_sys.plot_trajectory('xu')
cl_sys.plot_phase_plane_trajectory()
cl_sys.animate_simulation( time_factor_video=15.0 )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 30 14:04:51 2021

@author: alex
"""

import numpy as np

from pyro.dynamic.pendulum                import DoublePendulum
from pyro.planning.trajectoryoptimisation import DirectCollocationTrajectoryOptimisation


sys = DoublePendulum()

#Max/Min torque
sys.u_ub[0] = +20
sys.u_ub[1] = +20
sys.u_lb[0] = -20
sys.u_lb[1] = -20

planner = DirectCollocationTrajectoryOptimisation( sys , 0.2 , 20 )

planner.x_start = np.array([3.14,0,0,0])
planner.x_goal  = np.array([0,0,0,0])

planner.maxiter = 500
planner.set_linear_initial_guest(True)
planner.init_dynamic_plot()
planner.compute_optimal_trajectory()
# planner.show_solution()
planner.animate_solution()


###############################################################################
import numpy as np
import matplotlib.pyplot as plt
###############################################################################
from asimov import Asimov
from pyro.control import robotcontrollers
###############################################################################
kp = 80
ki = 0
kd = 0

asimov = Asimov()  # Asimov

asimov.x0 = np.array([-np.pi/4, -3*np.pi/4, np.pi/2, 0, 0, 0])  # Position initiale

qd = np.array([0.5, -np.pi/4, 0.5])  # Cible au joint
rd = asimov.forward_kinematic_effector(qd)

ctl = robotcontrollers.EndEffectorPID(asimov, kp, ki, kd)  # Déclaration du controlleur
ctl.rbar = rd  # Cible
ctl.kd = np.array([10, 30, 30])

closed_loop_robot = ctl + asimov  # Système boucle fermé

closed_loop_robot.plot_trajectory()  # Calcul de la trajectoire

closed_loop_robot.animate_simulation( is_3d = True )  # Animation et enregistrement




# -*- coding: utf-8 -*-
"""
Created on Jun 2 2021

@author: Alex
"""

from pyro.dynamic import massspringdamper
from pyro.control import linear


# Plant
m = 1
b = 1

plant = massspringdamper.FloatingSingleMass(m,b)

#Controller
kp  = 2
ki  = 0
kd  = 0
tau = 0.01

pid = linear.PIDController(kp,ki,kd,tau)

pid.rbar[0] = 2 # set point


closed_loop_sys = pid + plant

#closed_loop_sys.compute_trajectory( tf = 30 )
#closed_loop_sys.plot_trajectory('y')
#closed_loop_sys.animate_simulation()

closed_loop_sys.compute_trajectory( tf = 10 , solver = 'odeint' )
closed_loop_sys.plot_trajectory('xu')
#closed_loop_sys.animate_simulation()


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.dynamic  import manipulator
###############################################################################

torque_controlled_robot      = manipulator.TwoLinkManipulator()

# Target
q_desired = np.array([0.5,0.5])
r_desired = torque_controlled_robot.forward_kinematic_effector( q_desired )

# Joint PD

dof = 2

joint_PD      = robotcontrollers.JointPD( dof )
joint_PD.rbar = q_desired
joint_PD.kp   = np.array([25, 5 ])
joint_PD.kd   = np.array([ 1, 0 ])


# Effector PD 

model = torque_controlled_robot

effector_PD      = robotcontrollers.EndEffectorPD( model )
effector_PD.rbar = r_desired
effector_PD.kp   = np.array([100, 100 ])
effector_PD.kd   = np.array([  0,   0 ])

# Closed-loops

robot_with_joint_PD    = joint_PD    + torque_controlled_robot 
robot_with_effector_PD = effector_PD + torque_controlled_robot 

# Simulations

x0 = np.array([0,0,0,0])
tf = 5

anims = []

for robot in (robot_with_joint_PD, robot_with_effector_PD):
    robot.x0 = x0
    robot.compute_trajectory(tf)
    anims.append( robot.animate_simulation() )
    robot.plot_trajectory('xu')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 22 Feb 2024

@author: alex
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
import warnings

# Import standard graphical parameters if part of pyro
try:
    from pyro.analysis import graphical

    default_figsize = graphical.default_figsize
    default_dpi = graphical.default_dpi
    default_fontsize = graphical.default_fontsize
except:
    default_figsize = (10, 6)
    default_dpi = 100
    default_fontsize = 12


###############################################################################
class SingleAxisPolynomialTrajectoryGenerator:
    """
    This class is a tool to generate a point-to-point trajectory for a
    single axis based on boundary conditions (position and higher order derivative)

    Polynomial of order N

    x(t) = p0 + p1*t + p2*t^2 + ... + pN*t^N

    if boundary conditions do not fully specify the parameters of the polynomial,
    then an optimization is conducted to minimize the cost function which is defined
    as a weighted sum of the integral of the square of the ith derivative of the profile.

    Parameters:
    -----------
    tf : float
        duration of the trajectory
    poly_N : int
        order of the polynomial
    diff_N : int
        order of the highest derivative to compute
    x0 : array
        initial conditions (position, velocity, acceleration, jerk, snap, crackle, pop, ...)
    xf : array
        final conditions (position, velocity, acceleration, jerk, snap, crackle, pop, ...)
    x0_N : int
        number of initial conditions to impose (higher order derivative of the initial conditions)
    xf_N : int
        number of final conditions to impose (higher order derivative of the final conditions)
    Rs : array
        weights for the cost function penalizing the ith polynomial parameters directly
    Ws : array
        weights for the cost function penalizing the ith derivative of the profile
    dt : float
        time step for the numerical solution of the trajectory

    Output:
    -------
    p : array
        polynomial parameters
    X : array
        profile of the trajectory X[i, j] is the ith derivative of the profile at time t[j]
    t : array
        time vector

    """

    ################################################
    def __init__(
        self,
        tf=10,
        poly_N=5,
        diff_N=7,
        x0=np.array([0.0, 0.0, 0.0]),
        xf=np.array([0.0, 0.0, 10.0]),
        dt=0.01,
    ):

        self.tf = tf
        self.poly_N = poly_N
        self.diff_N = diff_N
        self.x0 = x0
        self.xf = xf
        self.x0_N = x0.shape[0]
        self.xf_N = xf.shape[0]
        self.Rs = np.zeros((self.poly_N + 1))
        self.Ws = np.zeros((self.diff_N))
        self.dt = dt

        # Outputs
        self.t = None
        self.X = None
        self.p = None

        self.labels = [
            "pos",
            "vel",
            "acc",
            "jerk",
            "snap",
            "crac",
            "pop",
            "7th",
            "8th",
            "9th",
            "10th",
        ]

    ################################################
    def compute_b(self, x0, xf, N0, Nf):
        """Compute the boundary condition vector b = [x0;xf] which represents the initial and final conditions on the trajectory and its derivatives"""

        b = np.hstack((x0[:N0], xf[:Nf]))

        print("Boundary condition vector b shape: ", b.shape)

        return b

    ################################################
    def compute_A(self, tf, N0, Nf, poly_N):
        """Compute the boundary condition matrix A which represents on the polynomial parameters are related to the boundary conditions"""

        A = np.zeros((N0 + Nf, poly_N + 1))

        # For all jth derivative of the initial conditions
        t0 = 0
        for j in range(N0):
            # For all terms of the polynomical
            for n in range(j, poly_N + 1):
                exp = n - j
                mul = 1
                for k in range(j):
                    mul = mul * (n - k)
                A[j, n] = mul * t0**exp

        # For all jth derivative of the final conditions
        for j in range(Nf):
            # For all terms of the polynomical
            for n in range(j, poly_N + 1):
                exp = n - j
                mul = 1
                for k in range(j):
                    mul = mul * (n - k)
                A[N0 + j, n] = mul * tf**exp

        print("Boundary condition matrix A shape: ", A.shape)

        return A

    ################################################
    def compute_Q(self, poly_N, diff_N, tf, Ws, Rs):
        """Compute the cost function matrix Q, only used if the boundary conditions do not fully specify the parameters of the polynomial"""

        # Quadratic cost matrix
        Q = np.zeros((poly_N + 1, poly_N + 1))

        # Quadratic cost matrix for each derivative
        Qs = np.zeros((poly_N + 1, poly_N + 1, diff_N))

        # Qs are weight corresponding to computing the integral of the square of the ith derivative of the profile
        # J = p.T @ Qs[i] @ p = integral( [ d_dt(ith)x(t) ]^2 dt)
        # see https://groups.csail.mit.edu/rrg/papers/BryIJRR15.pdf
        for r in range(diff_N):
            for i in range(poly_N + 1):
                for l in range(poly_N + 1):
                    if (i >= r) and (l >= r):
                        mul = 1
                        for m in range(r):
                            mul = mul * (i - m) * (l - m)
                        exp = i + l - 2 * r + 1
                        Qs[i, l, r] = 2 * mul * tf**exp / (i + l - 2 * r + 1)
                    else:
                        Qs[i, l, r] = 0

        # Total cost for all derivatives
        for r in range(diff_N):
            Q = Q + Ws[r] * Qs[:, :, r]

        # Regulation term penalizing the polynomial parameters directly
        Q = Q + np.diag(Rs[: (poly_N + 1)])

        return Q

    ################################################
    def solve_for_polynomial_parameters(self, A, b, Q):
        """Solve for the polynomial parameters pç

        Parameters:
        -----------
        A : array
            boundary condition matrix
        b : array
            boundary condition vector
        Q : array
            cost function matrix

        Output:
        -------
        p : array
            polynomial parameters

        """

        if A.shape[0] == A.shape[1]:

            print("Fully constrained trajectory parameters")
            p = np.linalg.solve(A, b)

        elif A.shape[0] > A.shape[1]:

            warnings.warn(
                "Warning! : impossible to respect all boundary condition, raise the order of the polynomial"
            )
            print(
                "Overconstrained boundary consitions: solving for best solution in the least-square sense"
            )
            p = np.linalg.lstsq(A, b)[0]

        else:

            print("Optimization over free decision variables")

            try:

                Q_inv = np.linalg.inv(Q)
                AQA = A @ Q_inv @ A.T
                lam = np.linalg.solve(AQA, b)
                p = Q_inv @ A.T @ lam

            except:

                print(" Q is not invertible, using optimization solver")

                p0 = np.zeros(A.shape[1])

                constraints = {"type": "eq", "fun": lambda p: A @ p - b}
                cost = lambda p: p.T @ Q @ p
                grad = lambda p: 2 * p.T @ Q
                hess = lambda p: 2 * Q

                # TODO: Change to a solver specifc to quadratic optimization
                res = minimize(
                    cost,
                    p0,
                    method="SLSQP",
                    jac=grad,
                    hess=hess,
                    constraints=constraints,
                    options={"disp": True, "maxiter": 5000},
                )

                p = res.x
            
            else:

                print(" Q is invertible, using direct solver")

        print("Computed polynomial parameters shape: ", p.shape)

        return p

    ################################################
    def generate_trajectory(self, tf, p, diff_N, dt=0.01):
        """Generate a numerical trajectory based on the polynomial parameters"""

        Np1 = p.shape[0]  # order of polynomial
        steps = int(tf / dt)  # number of time steps
        ts = np.linspace(0, tf, steps)
        X = np.zeros((diff_N, steps))

        # For all jth derivative of the signal
        for j in range(diff_N):
            # For all time steps
            for i in range(steps):
                t = ts[i]
                x = 0
                # For all terms of the polynomical
                # TODO could replace this with A(t) generic code
                for n in range(j, Np1):
                    p_n = p[n]
                    exp = n - j
                    mul = 1
                    for k in range(j):
                        mul = mul * (n - k)
                    x = x + mul * p_n * t**exp

                X[j, i] = x

        return X, ts

    ################################################
    def get_trajectory(self, j, t, p):
        """Get the jth derivative of the trajectory at time t based on the polynomial parameters p"""

        Np1 = p.shape[0]  # order of polynomial
        x = 0

        # For all terms of the polynomical
        for n in range(j, Np1):
            p_n = p[n]
            exp = n - j
            mul = 1
            for k in range(j):
                mul = mul * (n - k)
            x = x + mul * p_n * t**exp

        return x

    ################################################
    def plot_trajectory(self, X, t, n_fig=None):

        # Number of derivatives to plot
        n_max = X.shape[0]
        if n_fig is None:
            n = n_max
        elif n_fig < n_max:
            n = n_fig
        else:
            n = n_max

        fig, ax = plt.subplots(
            n,
            figsize=default_figsize,
            dpi=default_dpi,
            frameon=True,
        )

        if n == 1:
            ax = [ax]

        for i in range(n):

            ax[i].plot(t, X[i, :], "b")
            ax[i].set_ylabel(self.labels[i], fontsize=default_fontsize)
            ax[i].tick_params(labelsize=default_fontsize)
            ax[i].grid(True)

        ax[-1].set_xlabel("Time[sec]", fontsize=default_fontsize)
        # fig.tight_layout()
        fig.canvas.draw()

        plt.show()

    ################################################
    def solve(self, show=True):

        tf = self.tf
        x0 = self.x0
        xf = self.xf
        N0 = self.x0_N
        Nf = self.xf_N
        Np = self.poly_N
        Nd = self.diff_N
        Ws = self.Ws
        Rs = self.Rs
        dt = self.dt

        b = self.compute_b(x0, xf, N0, Nf)
        A = self.compute_A(tf, N0, Nf, Np)
        Q = self.compute_Q(Np, Nd, tf, Ws, Rs)

        p = self.solve_for_polynomial_parameters(A, b, Q)

        X, t = self.generate_trajectory(tf, p, Nd, dt)

        if show:
            self.plot_trajectory(X, t)

        return p, X, t


###############################################################################
class MultiPointSingleAxisPolynomialTrajectoryGenerator(
    SingleAxisPolynomialTrajectoryGenerator
):
    """
    This class is a tool to generate a point-to-point trajectory for a
    single axis based on boundary conditions (position and higher order derivative)

    k Polynomial segments of order N

    pi(t) = p0i + p1i*t + p2i*t^2 + ... + pNi*t^N

    x(T) = pi(t) with i s.t. ti < T < ti+1 and t = T - ti

    if boundary conditions do not fully specify the parameters of the polynomial,
    then an optimization is conducted to minimize the cost function which is defined
    as a weighted sum of the integral of the square of the ith derivative of the profile.

    Parameters:
    -----------
    poly_N : int
        order of the polynomials
    diff_N : int
        order of the highest derivative to compute
    tc : array
        array of time for initial, intermediate and final points
    x0 : array
        array of initial conditions
    xf : array
        array of final conditions
    xc : array ( # of constrained derivatives , pts_N)
        matrice of intermediate conditions at waypoints
    Rs : array
        weights for the cost function penalizing the ith polynomial parameters directly
    Ws : array
        weights for the cost function penalizing the ith derivative of the profile
    dt : float
        time step for the numerical solution of the trajectory

    Output:
    -------
    p : array
        polynomial parameters
    X : array
        profile of the trajectory X[i, j] is the ith derivative of the profile at time t[j]
    t : array
        time vector

    """

    ################################################
    def __init__(
        self,
        poly_N=5,
        diff_N=7,
        con_N=3,
        x0=np.array([0.0, 0.0, 0.0]),
        xf=np.array([10.0, 0.0, 0.0]),
        tc=np.array([0.0, 2.0, 8.0, 10.0]),
        xc=np.array([[3.0, 7.0], [0.0, 0.0]]),
        dt=0.01,
    ):
        self.poly_N = poly_N
        self.diff_N = diff_N
        self.x0 = x0
        self.xf = xf
        self.xc = xc
        self.tc = tc
        self.x0_N = x0.shape[0]
        self.xf_N = xf.shape[0]
        self.K = xc.shape[1]  # number of waypoints
        self.way_N = xc.shape[0]  # number of derivative to impose at each waypoint
        self.con_N = con_N  # number continuity constraints
        self.Rs = np.zeros((self.poly_N + 1))
        self.Ws = np.zeros((self.diff_N))
        self.dt = dt

        # Outputs
        self.t = None
        self.X = None
        self.p = None

        self.labels = [
            "pos",
            "vel",
            "acc",
            "jerk",
            "snap",
            "crac",
            "pop",
            "7th",
            "8th",
            "9th",
            "10th",
        ]

    ################################################
    def compute_b(self, x0, xf, xc, N0, Nf, Nw, Nc):
        """ """

        K = xc.shape[1] # number of waypoint

        b = x0[:N0]  # initial conditions

        for k in range(K):
            b = np.hstack((b, xc[:Nw, k]))  # waypoint conditions

        b = np.hstack((b, xf[:Nf]))

        for k in range(K):
            b = np.hstack((b, np.zeros(Nc)))  #  continuity conditions

        #print("Constraints vector b = \n", b)

        print("Constraints vector b shape = ",b.shape)

        return b
    
    ################################################
    def A_t(self, t, poly_N, diff_N):
        """ 
        Compute the matrix  X(t) = A(t) @ p
        where
        p = [p0, p1, p2, ..., pN] are the polynomial parameters
        X(t) = [x(t), dx(t), d2x(t), ..., dNx(t)] are the derivatives of the trajectory at time t

        """

        A = np.zeros((diff_N, poly_N + 1))

        # For all jth derivative of the final conditions
        for j in range(diff_N):
            # For all terms of the polynomical
            for n in range(j, poly_N + 1):
                exp = n - j
                mul = 1
                for k in range(j):
                    mul = mul * (n - k)
                A[j, n] = mul * t**exp

        return A


    ################################################
    def compute_A(self, tc, N0, Nf, Nw, Nc, Np):
        """"""
        Kp1 = tc.shape[0]-1  # number of segments
        K = Kp1 - 1 # number of waypoints
        N = Np + 1 # number of polynomial parameters per segments
        N_params = Kp1 * N

        N_contraints = N0 + Nf + Nw * K + Nc * K

        A = np.zeros((N_contraints, N_params))

        # Times on segments
        dt = np.diff(tc) # time between waypoints

        # Initial conditions
        A[0:N0, 0:N] = self.A_t(0, Np, N0)

        # Waypoint conditions
        for k in range(K):
            A[N0 + Nw * k : N0 + Nw * (k + 1), N * (k + 1) : N * (k + 2)] = self.A_t(0, Np, Nw)

        # Final conditions
        A[N0 + Nw * K : N0 + Nw * K + Nf, N * K : N * (K + 1)] = self.A_t(dt[-1], Np, Nf)

        # Continuity conditions
        for k in range(K):
            A[N0 + Nw * K + Nf + Nc * k : N0 + Nw * K + Nf + Nc * (k + 1), N * k : N * (k + 1)] = self.A_t(dt[k], Np, Nc)
            A[N0 + Nw * K + Nf + Nc * k : N0 + Nw * K + Nf + Nc * (k + 1), N * (k + 1) : N * (k + 2)] = -self.A_t(0, Np, Nc)

        #print("Constraints matrix: \n", A)

        print("Constraints matrix A shape = ",A.shape)

        return A
    
    ################################################
    def compute_Q(self, poly_N, diff_N, tc, Ws, Rs):

        #
        Kp1 = tc.shape[0]-1  # number of segments
        N = poly_N + 1 # number of polynomial parameters per segments
        N_params = Kp1 * N

        # Times on segments
        dt = np.diff(tc) # time between waypoints

        # Quadratic global cost matrix
        N_params = (poly_N + 1) * Kp1  # number of polynomial parameters
        Q = np.zeros((N_params, N_params))

        # Waypoint conditions
        for k in range(Kp1):
            Qk = self.compute_segment_Q(poly_N, diff_N, dt[k], Ws, Rs)
            Q[N * k : N * (k + 1), N * k : N * (k + 1)] = Qk

        #print("Quadratic cost matrix: \n", Q)

        print("Quadratic cost matrix shape = ",Q.shape)

        return Q


    ################################################
    def compute_segment_Q(self, poly_N, diff_N, tf, Ws, Rs):
        """Compute the cost function matrix Q, only used if the boundary conditions do not fully specify the parameters of the polynomial"""

        # Quadratic cost matrix
        Q = np.zeros((poly_N + 1, poly_N + 1))

        # Quadratic cost matrix for each derivative
        Qs = np.zeros((poly_N + 1, poly_N + 1, diff_N))

        # Qs are weight corresponding to computing the integral of the square of the ith derivative of the profile
        # J = p.T @ Qs[i] @ p = integral( [ d_dt(ith)x(t) ]^2 dt)
        # see https://groups.csail.mit.edu/rrg/papers/BryIJRR15.pdf
        for r in range(diff_N):
            for i in range(poly_N + 1):
                for l in range(poly_N + 1):
                    if (i >= r) and (l >= r):
                        mul = 1
                        for m in range(r):
                            mul = mul * (i - m) * (l - m)
                        exp = i + l - 2 * r + 1
                        Qs[i, l, r] = 2 * mul * tf**exp / (i + l - 2 * r + 1)
                    else:
                        Qs[i, l, r] = 0

        # Total cost for all derivatives
        for r in range(diff_N):
            Q = Q + Ws[r] * Qs[:, :, r]

        # Regulation term penalizing the polynomial parameters directly
        Q = Q + np.diag(Rs[: (poly_N + 1)])

        return Q

    ################################################
    def generate_trajectory(self, tc, p, poly_N, diff_N, dt=0.01):
        """Generate a numerical trajectory based on the polynomial parameters"""

        Np1 = poly_N + 1  # order of polynomial
        tf =tc[-1]
        steps = int(tf / dt)  # number of time steps
        ts = np.linspace(0, tf, steps)
        X = np.zeros((diff_N, steps))

        for i in range(steps):
            t = ts[i]
            k = 0
            t0_segment = 0.0
            while (t > tc[k+1]):
                t0_segment = tc[k+1]
                k = k + 1

            t_local = t - t0_segment 
            p_local = p[Np1 * k : Np1 * (k + 1)]
            At = self.A_t(t_local, Np1 - 1, diff_N)
            X[:, i] = At @ p_local

        return X, ts

    ################################################
    def get_local_trajectory(self, t_local, p_local, diff_N):
        """Get the jth derivative of the trajectory at time t based on the polynomial parameters p"""

        p = p_local
        Np = p.shape[0] - 1  # order of polynomial
        t = t_local
        At = self.A_t(t, Np, diff_N)

        X = At @ p

        return X

    ################################################
    def solve(self, show=True):

        x0 = self.x0
        xf = self.xf
        xc = self.xc
        tc = self.tc
        N0 = self.x0_N
        Nf = self.xf_N
        Nw = self.way_N
        Nc = self.con_N
        Np = self.poly_N
        Nd = self.diff_N
        Ws = self.Ws
        Rs = self.Rs
        dt = self.dt

        b = self.compute_b(x0, xf, xc, N0, Nf, Nw, Nc)
        A = self.compute_A(tc, N0, Nf, Nw, Nc, Np) 
        Q = self.compute_Q(Np, Nd, tc, Ws, Rs) 

        p = self.solve_for_polynomial_parameters(A, b, Q)

        X, t = self.generate_trajectory(tc, p, Np, Nd, dt)

        if show:
            self.plot_trajectory(X, t)

        return b, A, p, X, t


"""
#################################################################
##################          Main                         ########
#################################################################
"""


if __name__ == "__main__":
    """MAIN TEST"""

    x0 = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    xf = np.array([10, 0, 0, 0, 0, 0, 0, 0])

    ge = SingleAxisPolynomialTrajectoryGenerator(
        x0=x0, xf=xf, tf=10, poly_N=12, diff_N=7, dt=0.01
    )

    #############################
    ### Fully constrained order 3
    #############################

    # ge.x0_N = 2
    # ge.xf_N = 2
    # ge.poly_N = 3
    # ge.diff_N = 7

    # ge.solve()

    #############################
    ### Fully constrained order 5
    #############################

    # ge.x0_N = 3
    # ge.xf_N = 3
    # ge.poly_N = 5
    # ge.diff_N = 7

    # ge.solve()  # order 5 fully constrained

    ###########################################
    ### Optimization on polynomial parameters
    ###########################################

    ge.poly_N = 12
    ge.x0_N = 3
    ge.xf_N = 3
    ge.Rs = 0.0 * np.ones(ge.poly_N + 1)
    # ge.Ws = np.array([0, 0.0, 10.0, 1.0, 1.0, 1.0, 1.0])
    ge.Ws = np.array([0, 1.0, 0.0, 0, 0.0, 0, 0])
    # ge.Ws = np.array([0.01, 10.0, 0.01, 0.01, 0.01, 0.01, 0.01])

    p, X, t = ge.solve()  # order 12 with optimization on polynomial parameters

    #############################
    ### Fully constrained order 7
    #############################

    # ge = SingleAxisPolynomialTrajectoryGenerator(
    #     x0=x0, xf=xf, tf=10, poly_N=7, diff_N=7, dt=0.01
    # )

    # ge.x0_N = 4
    # ge.xf_N = 4

    # ge.solve()

    #############################
    ### Fully constrained order 9
    #############################

    ge = SingleAxisPolynomialTrajectoryGenerator(
        x0=x0, xf=xf, tf=10, poly_N=9, diff_N=7, dt=0.01
    )

    ge.x0_N = 5
    ge.xf_N = 5

    ge.solve()

    #############################
    ### Overconstrained order 3
    #############################

    # ge = SingleAxisPolynomialTrajectoryGenerator(
    #     x0=x0, xf=xf, tf=10, poly_N=3, diff_N=7, dt=0.01
    # )

    # ge.x0_N = 3
    # ge.xf_N = 3

    # ge.solve()

    #############################
    ### Waypoint simple fully constraint test
    #############################

    # traj = MultiPointSingleAxisPolynomialTrajectoryGenerator(
    #     poly_N=3,
    #     diff_N=5,
    #     con_N=2,
    #     x0=np.array([0.0, 0.0]),
    #     xf=np.array([10.0, 0.0]),
    #     tc=np.array([0.0, 2.0, 8.0, 10.0]),
    #     xc=np.array([[3.0, 7.0], [1.0, 1.0]]),
    #     dt=0.01,
    # )

    # b, A, p, X, t = traj.solve()

    #############################
    ### Waypoint test
    #############################

    # traj = MultiPointSingleAxisPolynomialTrajectoryGenerator(
    #     poly_N=3,
    #     diff_N=5,
    #     con_N=3,
    #     x0=np.array([0.0, 0.0]),
    #     xf=np.array([10.0, 0.0]),
    #     tc=np.array([0.0, 2.0, 8.0, 10.0]),
    #     xc=np.array([[3.0, 7.0]]),
    #     dt=0.01,
    # )

    # b, A, p, X, t = traj.solve()

    #############################
    ### Waypoint test
    #############################

    # traj = MultiPointSingleAxisPolynomialTrajectoryGenerator(
    #     poly_N=4,
    #     diff_N=5,
    #     con_N=4,
    #     x0=np.array([0.0, 0.0, 0.0]),
    #     xf=np.array([10.0, 0.0]),
    #     tc=np.array([0.0, 2.0, 8.0, 10.0]),
    #     xc=np.array([[3.0, 7.0]]),
    #     dt=0.01,
    # )

    # b, A, p, X, t = traj.solve()

    #############################
    ### Waypoint simple fully constraint test
    #############################

    traj = MultiPointSingleAxisPolynomialTrajectoryGenerator(
        poly_N=9,
        diff_N=7,
        con_N=4,
        x0=np.array([0.0, 0.0, 0.0]),
        xf=np.array([10.0, 0.0, 0.0]),
        tc=np.array([0.0, 2.0, 8.0, 10.0]),
        xc=np.array([[5.0, 12.0]]),
        dt=0.01,
    )

    traj.Ws[0]= 0.01
    traj.Ws[1]= 1.0
    traj.Ws[2]= 1.0
    traj.Ws[3]= 1.0
    traj.Ws[4]= 1.0

    b, A, p, X, t = traj.solve()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:27:47 2022

@author: alex
"""

import numpy as np

from pyro.dynamic  import pendulum
from pyro.control  import nonlinear
from pyro.analysis import costfunction
from pyro.planning import dynamicprogramming 
from pyro.planning import discretizer

sys  = pendulum.SinglePendulum()

sys.x_ub = np.array([+6, +6])
sys.x_lb = np.array([-9,  -6])


# Get optimal ctl
ctl = nonlinear.ComputedTorqueController( sys )
ctl.rbar = np.array([-3.14]) # target

# Cost Function
qcf = costfunction.QuadraticCostFunction.from_sys(sys)

qcf.xbar = np.array([ ctl.rbar[0] , 0 ]) # target
qcf.INF  = 300

qcf.S[0,0] = 10.0
qcf.S[1,1] = 10.0

# Min/max torque
sys.u_ub[0] = +200
sys.u_lb[0] = -200

# Evaluate on a grid
grid_sys = discretizer.GridDynamicSystem( sys , [301,301] , [11] , 0.05 , False )

evaluator = dynamicprogramming.PolicyEvaluatorWithLookUpTable(ctl, grid_sys, qcf)
evaluator.solve_bellman_equation()
evaluator.plot_cost2go()



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 12 20:13:30 2019

@author: alxgr
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.dynamic import cartpole
from pyro.control import controller
###############################################################################

class CartPoleController( controller.StaticController ) :
    
    ############################
    def __init__( self , sys_model ):
        """ """
        
        # Dimensions
        self.dof = 2
        self.k   = 1 
        self.m   = 1
        self.p   = 4 
        
        super().__init__(self.k, self.m, self.p)
        
        # Label
        self.name = 'Cart Pole Controller'
        
        # Gains
        self.kd = 5
        self.kp = 3
        
        self.rbar = np.zeros(1)
        
        
        # Model
        self.sys = sys_model

    
    #############################
    def c( self , y , r , t = 0 ):
        """ 
        Feedback static computation u = c(y,r,t)
        
        INPUTS
        y  : sensor signal_proc vector     p x 1
        r  : reference signal_proc vector  k x 1
        t  : time                     1 x 1
        
        OUPUTS
        u  : control inputs vector    m x 1
        
        """
        
        u = np.zeros(self.m) 
        
        # State feedback
        q  = y[ 0        :     self.dof   ]
        dq = y[ self.dof : 2 * self.dof   ]
        
        H = self.sys.H(q)
        C = self.sys.C(q,dq)
        d = self.sys.d(q,dq)
        g = self.sys.g(q)
        
        phi = np.dot(C,dq) + d + g
        
        H11 = H[0,0]
        H12 = H[0,1]
        H22 = H[1,1]
        
        ddq_r = - self.kd * dq[1] - self.kp * q[1]
        
        if ( np.abs(H12) > 0.001 ):
            
            A = H12 - H11 * H22 / H12
            B = phi[0] - H11 / H12 * phi[1]
            
            u[0] = A * ddq_r + B 
            
        else:
            
            u[0] = 0

        
        return u



###############################################################################

sys = cartpole.UnderActuatedRotatingCartPole()

ctl = CartPoleController( sys )

# New cl-dynamic
cl_sys = ctl + sys

# Simultation
cl_sys.x0 = np.array([0,-3.14,0,0])
cl_sys.compute_trajectory( 10 )
cl_sys.plot_trajectory('xu')
cl_sys.animate_simulation(time_factor_video=1.0, is_3d=True)


# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 10:17:36 2018

@author: Alexandre
"""

###############################################################################
import numpy as np

###############################################################################
from pyro.dynamic  import integrator
from pyro.analysis import costfunction
    
###################################
# Simple integrator
###################################

sys = integrator.DoubleIntegrator()

sys.ubar = np.array([1]) # constant input = 1


###################################
# Analysis
###################################
    
# Phase plane behavior test
sys.plot_phase_plane()

# Simulation
sys.x0 = np.array([0,0])

traj = sys.compute_trajectory( tf = 10 )

sys.plot_trajectory()
sys.plot_trajectory('y')
sys.plot_phase_plane_trajectory()

# Cost computing

qcf           = costfunction.QuadraticCostFunction(2, 1)
traj_with_qcf = qcf.trajectory_evaluation( traj )
sys.traj      = traj_with_qcf
sys.plot_trajectory('xuj')

tcf           = costfunction.TimeCostFunction( np.array([0,0]) )
traj_with_tcf = tcf.trajectory_evaluation( traj )
sys.traj      = traj_with_tcf
sys.plot_trajectory('xuj')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 24 13:05:15 2020

@author: alex
"""

##############################################################################
import numpy as np
from scipy.linalg  import solve_continuous_are

##############################################################################
from pyro.control  import linear
from pyro.control  import controller
from pyro.dynamic  import statespace
from pyro.analysis import costfunction
##############################################################################


#################################################################
def synthesize_lqr_controller( ss , cf , xbar = None, ubar = None):
    """

    Compute the optimal linear controller minimizing the quadratic cost:
        
    J = int ( xQx + uRu ) dt = xSx
    
    with control law:
        
    u = K x
    
    Note:
    ---------
    Controller assume y = x  (output is directly the state vector)

    Parameters
    ----------
    sys : `StateSpaceSystem` instance
    cf  : "quadratic cost function" instance
    xbar: offset on state feedback
    ubar: offset on control input
        
    Returns
    -------
    instance of `Proportionnal Controller`

    """
    
    # Solve Riccati Equation
    # A’X + XA - XBR^-1B’X+Q=0
    S = solve_continuous_are( ss.A , ss.B , cf.Q , cf.R )
    
    # Matrix gain
    BTS   = np.dot( ss.B.T , S )
    R_inv = np.linalg.inv( cf.R )
    K     = np.dot( R_inv  , BTS )
    
    # Define linear controller
    ctl = linear.ProportionalController.from_matrix( K )
    ctl.name = 'LQR controller'
    
    # Offsets on input and ouputs
    if xbar is not None:
        ctl.ybar = xbar  # Offset on the sensor signal
        
    if ubar is not None:
        ctl.ubar = ubar  # Offset on the control input
    
    
    return ctl


#################################################################
def linearize_and_synthesize_lqr_controller( sys , cf ):
    """

    Compute the optimal linear controller minimizing the quadratic cost:
        
    J = int ( xQx + uRu ) dt = xSx
    
    with control law:
        
    u = K x
    
    Note:
    ---------
    Controller assume y = x  (output is directly the state vector)

    Parameters
    ----------
    sys : `ContinuousSystem` instance
    cf  : "quadratic cost function" instance
        
    Returns
    -------
    instance of `Proportionnal Controller`

    """
    
    ss  = statespace.linearize( sys , 0.01 )
    
    ctl = synthesize_lqr_controller( ss , cf , sys.xbar , sys.ubar )
    
    return ctl



##############################################################################
### Time varying LQR controller
##############################################################################

class TrajectoryLQRController( controller.StaticController ):
    """
    General (SISO or MIMO) proportional controller
    -------------------------------------------------

    u = u_d(t) + K(t) ( xd(t) - x )
    
    
    K = R^(-1) B(t) S(t)
    
    S(t) = integral of differential riccati equation
    
    A(t) = df/dx(t)
    B(t) = df/du(t)
    
    x = y  : state feedback is assumed
    
    -----------------------------------------
    r  : reference signal vector       k x 1
    y  : sensor signal vector          p x 1
    u  : control inputs vector         m x 1
    -----------------------------------------
    
    """
    
    ###############################
    def __init__(self, sys , traj , cf = None ):
        
        n = sys.n # states dimensions
        m = sys.m # input dimenstions
        
        controller.StaticController.__init__( self, n, m, n)
        
        self.name = "Trajectory LQR Controller"
        
        self.sys  = sys
        self.traj = traj
        
        if cf is None:
            
            self.cf = sys.cost_function
            
        else:
            
            self.cf = cf
            
        self.compute_linear_dynamic()
        self.compute_cost_matrices()
        self.compute_cost2go()
        self.compute_gains()
        
        self.traj.generate_interpol_functions()
        
        print('\n\n-------------------------------------------------------------------------')
        print('Warning: this method is an approximation of the riccati solution for now')
        print('ToDo @Alex : fix differential riccati integration')
        print('-------------------------------------------------------------------------\n')
            
            
    ###############################
    def compute_linear_dynamic(self):
        
        steps = self.traj.time_steps
        
        self.As = np.zeros(( self.sys.n , self.sys.n , steps ))
        self.Bs = np.zeros(( self.sys.n , self.sys.m , steps ))
        
        for i in range(steps):
            
            self.sys.tbar = self.traj.t[i]
            self.sys.xbar = self.traj.x[i,:]
            self.sys.ubar = self.traj.u[i,:]
            
            # TODO more intelligent gradient computation
            ss  = statespace.linearize( self.sys , 0.01 )
            
            self.As[:,:,i] = ss.A
            self.Bs[:,:,i] = ss.B
            
            
    
    ###############################
    def compute_cost_matrices(self):
        
        steps = self.traj.time_steps
        
        self.Qs = np.zeros(( self.sys.n , self.sys.n , steps ))
        self.Rs = np.zeros(( self.sys.m , self.sys.m , steps ))
        
        for i in range(steps):
            
            # ToDo Add Time Varying Quadratic cost object
            self.Qs[:,:,i] = self.cf.Q
            self.Rs[:,:,i] = self.cf.R
        
        
        
    ###############################
    def compute_cost2go(self):
        
        steps = self.traj.time_steps
        
        self.Ss = np.zeros(( self.sys.n , self.sys.n , steps ))
        self.Ps = np.zeros(( self.sys.n , self.sys.n , steps ))
        
        # Boundary conditions
        Af = self.As[:,:,-1]
        Bf = self.Bs[:,:,-1]
        Qf = self.Qs[:,:,-1]
        Rf = self.Rs[:,:,-1]
        
        self.Ss[:,:,-1] = solve_continuous_are( Af , Bf , Qf , Rf )
        
        # self.Ss[:,:,-1] = np.eye( self.sys.n )
        # self.Ps[:,:,-1] = np.eye( self.sys.n )
        
        for i in range( steps -1 , 0, -1 ):
            
            A = self.As[:,:,i]
            B = self.Bs[:,:,i]
            Q = self.Qs[:,:,i]
            R = self.Rs[:,:,i]
            S = self.Ss[:,:,i]
            # P = self.Ps[:,:,i]
            
            # Quick approx for testing
            dS = Q 
            
            ##############
            ## TODO: Make a real integraiton of differential riccati equation
            #############
            
            # dS = S @ A + A.T @ S - S @ B @ np.linalg.inv(R) @ B.T @ S + Q
            # dP = A.T @ P - 0.5 * S @ B @ np.linalg.inv(R) @ B.T @ P + 0.5 * Q @ np.linalg.inv(P).T
            
            dt = self.traj.t[i] - self.traj.t[i-1]
            
            # Euler integration (maybe is not engouh precision)
            self.Ss[:,:,i-1] = self.Ss[:,:,i] + dS * dt
            
            # self.Ps[:,:,i-1] = self.Ps[:,:,i] + dP * dt
            # self.Ss[:,:,i-1] = self.Ps[:,:,i-1] @ self.Ps[:,:,i-1].T
            
            # print('t = ', self.traj.t[i])
            # print('\ndS = \n' , dS)
            # print('\nS = \n' , S)
            
            
    ###############################
    def compute_gains(self):
        
        steps = self.traj.time_steps
        
        self.Ks = np.zeros(( self.sys.m , self.sys.n , steps ))
        
        for i in range( steps ):
            
            A = self.As[:,:,i]
            B = self.Bs[:,:,i]
            Q = self.Qs[:,:,i]
            R = self.Rs[:,:,i]
            S = self.Ss[:,:,i]
            
            K = np.linalg.inv(R) @ B.T @  S
            
            print('\n-------------------\nt = ', self.traj.t[i])
            print('K = \n' , K)
            
            self.Ks[:,:,i] = K
        
        
    ##############################
    def c(self, x, r, t):
        """ Feedback law """
        
        if t < self.traj.time_final:
            
            # Nominal control input
            u_bar = self.traj.inter_t2u( t )
            x_bar = self.traj.inter_t2x( t )
            
            # Find closest traj point
            i = (np.abs(self.traj.t - t)).argmin()
            
            K = self.Ks[:,:,i]  # Time varying linear gain
            
            u_e = K @ ( x_bar - x )
            
            u = u_bar + u_e
            
        else:
            
            x_bar = self.traj.x[-1,:]
            
            K = self.Ks[:,:,-1]
            
            u = K @ ( x_bar - x ) # this assume final position is a fixed point
        
        return u


    


'''
#################################################################
##################          Main                         ########
#################################################################
'''


if __name__ == "__main__":
    
    
    from pyro.dynamic.pendulum      import DoublePendulum
    from pyro.analysis.costfunction import QuadraticCostFunction
    
    sys = DoublePendulum()
    
    ss  = statespace.linearize( sys , 0.01 )
    
    cf  = QuadraticCostFunction.from_sys( sys )
    
    # Tune cost function here:
    cf.R[0,0] = 1000
    cf.R[1,1] = 10000
    
    ctl = synthesize_lqr_controller( ss , cf )
    
    print('\nLinearized sys matrix A:')
    print(ss.A)
    print('\nLinearized sys matrix B:')
    print(ss.B)
    
    print('\nCost function matrix Q:')
    print(cf.Q)
    print('\nCost function matrix R:')
    print(cf.R)
    
    print('\nLQR Controller gain K:')
    print(ctl.K)
    
    
    # Simulation Open-Loop Non-linear
    sys.x0 = np.array([0.1,0.2,0,0])
    sys.plot_trajectory()
    sys.animate_simulation()
    
    # Simulation Open-Loop Linearized
    ss.x0 = np.array([0.1,0.2,0,0])
    ss.compute_trajectory()
    sys.traj = ss.traj
    sys.animate_simulation()
    
    # Simulation Closed-Loop Non-linear with LQR controller
    cl_sys = ctl + sys
    cl_sys.x0 = np.array([0.1,0.2,0,0])
    cl_sys.compute_trajectory()
    cl_sys.plot_trajectory('xu')
    cl_sys.animate_simulation()
    


# -*- coding: utf-8 -*-
"""
Created on Sun May 12 16:34:44 2019

@author: Alexandre
"""

###############################################################################
import numpy as np
###############################################################################
from pyro.control  import robotcontrollers
from pyro.dynamic  import manipulator
###############################################################################

# Dynamic model
torque_controlled_robot = manipulator.FiveLinkPlanarManipulator()
sys  = manipulator.SpeedControlledManipulator.from_manipulator( torque_controlled_robot )

# Controller
ctl = robotcontrollers.EndEffectorKinematicControllerWithNullSpaceTask( sys )

# Main objective
ctl.rbar  = np.array([1.0,1.0])
ctl.gains = np.array([1.0,1.0])

# Secondary objective
ctl.q_d         = np.array([-1,-1,-1,-1,-1])
ctl.gains_null  = np.array([10,10,10,10,10])


# Closed-loop dynamics
cl_sys = ctl + sys

# Simulation and plots
cl_sys.x0        = np.array([0.1,0.1,0.1,0.1,0.1])

cl_sys.plot_trajectory('x')
cl_sys.animate_simulation()



# -*- coding: utf-8 -*-

###############################################################################
import numpy as np
import matplotlib.pyplot as plt

from pyro.analysis import graphical
from pyro.planning import trajectorygeneration
from pyro.dynamic import drone
from pyro.planning import plan

###############################################################################
# Define the c.g. trajectory
###############################################################################

# Waypoints       x_i   y_i  t_i
xyt = np.array([[ 0.0,  0.0, 0.0],
                [10.0,  0.0, 5.0], 
                [10.0, 15.0, 10.0], 
                [ 0.0, 5.0, 15.0]]).T



traj = trajectorygeneration.MultiPointSingleAxisPolynomialTrajectoryGenerator(
        poly_N=9,
        diff_N=7,
        con_N=4,
        x0=np.array([xyt[0,0], 0.0, 0.0, 0.0]),
        xf=np.array([xyt[0,3], 0.0, 0.0, 0.0]),
        tc=np.array([xyt[2,0], xyt[2,1], xyt[2,2], xyt[2,3]]),
        xc=np.array([[xyt[0,1], xyt[0,2]]]),
        dt=0.01,
    )

traj.Ws[0]= 0.01
traj.Ws[1]= 1.0
traj.Ws[2]= 1.0
traj.Ws[3]= 1.0
traj.Ws[4]= 1.0

b, A, p, X, t = traj.solve()

x = X[0, :]
dx = X[1, :]
ax = X[2, :]
dax = X[3, :]
ddax = X[4, :]

traj = trajectorygeneration.MultiPointSingleAxisPolynomialTrajectoryGenerator(
        poly_N=9,
        diff_N=7,
        con_N=4,
        x0=np.array([xyt[1,0], 0.0, 0.0, 0.0]),
        xf=np.array([xyt[1,3], 0.0, 0.0, 0.0]),
        tc=np.array([xyt[2,0], xyt[2,1], xyt[2,2], xyt[2,3]]),
        xc=np.array([[xyt[1,1], xyt[1,2]]]),
        dt=0.01,
    )

traj.Ws[0]= 0.01
traj.Ws[1]= 1.0
traj.Ws[2]= 1.0
traj.Ws[3]= 1.0
traj.Ws[4]= 1.0

b, A, p, Y, t = traj.solve()

y = Y[0, :]
dy = Y[1, :]
ay = Y[2, :] + 9.81  # added gravity
day = Y[3, :]
dday = Y[4, :]

###############################################################################
# Compute angular trajectory
###############################################################################

theta = np.arctan2(ax, ay)
# theta = np.arctan( (ay/ax))
s = np.sin(theta)
c = np.cos(theta)
dtheta = (dax * c - day * s) / (ax * s + ay * c)
ddtheta = (
    s * (-dday + ay * dtheta**2 - 2 * dax * dtheta)
    + c * (ddax - ax * dtheta**2 - 2 * day * dtheta)
) / (ay * c + ax * s)

###############################################################################
# Compute Forces
###############################################################################

sys = drone.Drone2D()

m = sys.mass
J = sys.inertia
r = sys.truster_offset

F_tot = m * ax / np.sin(theta)
F_del = J * ddtheta / r

T1 = 0.5 * (F_tot + F_del)
T2 = 0.5 * (F_tot - F_del)

###############################################################################
# Plots
###############################################################################

# Create traj
steps = len(t)
us = np.zeros((steps, 2))  # control inputs
xs = np.zeros((steps, 6))
ys = np.zeros((steps, 6))
dxs = np.zeros((steps, 6))

xs[:, 0] = x
xs[:, 1] = y
xs[:, 2] = -theta
xs[:, 3] = dx
xs[:, 4] = dy
xs[:, 5] = -dtheta

us[:, 0] = T1
us[:, 1] = T2

traj = plan.Trajectory(xs, us, t, dxs, ys)


fig, axes = plt.subplots(
    3, figsize=graphical.default_figsize, dpi=graphical.default_dpi, frameon=True
)

axes[0].plot(t, theta, "b")
axes[0].set_ylabel("Theta", fontsize=graphical.default_fontsize)
axes[0].set_xlabel("v", fontsize=graphical.default_fontsize)
axes[0].tick_params(labelsize=graphical.default_fontsize)
axes[0].grid(True)

axes[1].plot(t, dtheta, "b")
# axes[1].plot(t, dtheta_num, "r")
axes[1].set_ylabel("w", fontsize=graphical.default_fontsize)
axes[1].set_xlabel("t", fontsize=graphical.default_fontsize)
axes[1].tick_params(labelsize=graphical.default_fontsize)
axes[1].grid(True)

axes[2].plot(t, ddtheta, "b")
# axes[2].plot(t, ddtheta_num, "r")
axes[2].set_ylabel("dw", fontsize=graphical.default_fontsize)
axes[2].set_xlabel("t", fontsize=graphical.default_fontsize)
axes[2].tick_params(labelsize=graphical.default_fontsize)
axes[2].grid(True)

fig.tight_layout()
fig.canvas.draw()

plt.show()


fig, axes = plt.subplots(
    1, figsize=graphical.default_figsize, dpi=graphical.default_dpi, frameon=True
)

axes.plot(x, y, "r")
axes.set_ylabel("y", fontsize=graphical.default_fontsize)
axes.set_xlabel("x", fontsize=graphical.default_fontsize)
axes.axis("equal")
axes.set(xlim=(-25, 25), ylim=(-25, 25))
axes.tick_params(labelsize=graphical.default_fontsize)
axes.grid(True)

fig.tight_layout()
fig.canvas.draw()

plt.show()


sys.traj = traj

sys.plot_trajectory("xu")
sys.animate_simulation()
